/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file        scxcondition.cpp

    \brief       Condition support for SCX.
                 Also helps keep work+sleep time constant in a loop
    
    \date        12-18-2008 15:54:00

*/
/*----------------------------------------------------------------------------*/

#include <scxcorelib/scxcmn.h>
#include <scxcorelib/scxcondition.h>
//#include <scxcorelib/scxthread.h>
#include <scxcorelib/scxlog.h>
#include <scxcorelib/stringaid.h>

#if defined(sun) || defined(linux) || defined(hpux) || defined(aix)
    #include <errno.h>
    #include <time.h>
#elif defined(macos)
    #include <errno.h>
    #include <sys/time.h>    // for gettimeofday()
#elif defined(WIN32)
    #include <Windows.h>
#else
    #error Platform not implemented
#endif

using namespace SCXCoreLib;

namespace SCXCoreLib
{
/*----------------------------------------------------------------------------*/
/**
    Constructor

    \throws      SCXInternalErrorException if too large sleep time specified
    \throws      SCXErrnoException if error occurred calling system function
*/
    SCXCondition::SCXCondition()
        : m_bSetCalled(false)
          , m_sleepTime(0)
#if defined(SCX_UNIX)
          , m_bRecomputeSleep(false)
#elif defined(WIN32)
          , m_timeStamp(0)
#endif
    {
#if defined(SCX_UNIX)
        // Initialize the time specification (this is ASSERTed later in code!)
        memset(&m_ts, 0, sizeof(m_ts));

        int err;
        if (0 != (err = pthread_mutex_init(&m_lock, NULL)))
        {
            throw SCXErrnoException(L"pthread_mutex_init() function call failed", err, SCXSRCLOCATION);
        }

        if (0 != (err = pthread_cond_init(&m_cond, NULL)))
        {
            throw SCXErrnoException(L"pthread_cond_init() function call failed", err, SCXSRCLOCATION);
        }
#elif defined(WIN32)
        InitializeCriticalSection(&m_lock);

        if (NULL == (m_cond = CreateEvent(NULL, true, false, NULL)))
        {
            throw SCXErrnoException(L"CreateEvent() function failed", GetLastError(), SCXSRCLOCATION);
        }
#else
        #error Platform not supported
#endif
    }

/*----------------------------------------------------------------------------*/
/**
    Destructor

    \throws      SCXInternalErrorException if we are inside a condition test
    \throws      SCXErrnoException if error occurred calling system function
*/
    SCXCondition::~SCXCondition()
    {
#if defined(SCX_UNIX)
        int err;
        if (0 != (err = pthread_cond_destroy(&m_cond)))
        {
            throw SCXErrnoException(L"pthread_cond_destroy() function call failed", err, SCXSRCLOCATION);
        }

        if (0 != (err = pthread_mutex_destroy(&m_lock)))
        {
            throw SCXErrnoException(L"pthread_mutex_destroy() function call failed", err, SCXSRCLOCATION);
        }
#elif defined(WIN32)
                DeleteCriticalSection(&m_lock);

                if (0 == CloseHandle(m_cond))
                {
                        throw SCXErrnoException(L"CloseHandle() function failed", GetLastError(), SCXSRCLOCATION);
                }
#else
                #error Platform not supported
#endif
    }


/*----------------------------------------------------------------------------*/
/**
    Call before entering condition to set the starting timestamp and sleep time

    \param       milliseconds    Number of milliseconds for work+sleep

    \throws      SCXInternalErrorException if called multiple times
*/
    void SCXCondition::SetSleep(scxulong milliseconds)
    {
        if (!m_bSetCalled)
        {
            m_sleepTime = milliseconds;

#if defined(SCX_UNIX)
            GetMillisecondTimeStamp();

            if (m_sleepTime)
            {
                m_bRecomputeSleep = true;
            }
#elif defined(WIN32)
            m_timeStamp = GetMillisecondTimeStamp();
#endif

            m_bSetCalled = true;
        }
        else
        {
            throw SCXInternalErrorException(L"Can only call SetSleep() once", SCXSRCLOCATION);
        }
    }

/*----------------------------------------------------------------------------*/
/**
    Begin a condition test

    \throws      SCXInternalErrorException if we are already in a condition
*/
    void SCXCondition::BeginCondition()
    {
        // Lock our mutex
        Lock();
    }

/*----------------------------------------------------------------------------*/
/**
    Waits for a condition

    Sleeps as directed (sleep time passed in constructor).  If enough time has
    passed since last sleep where we should be running again, then sleep is not
    performed (we wake up immediately).

    This is "automatic" on UNIX/Linux because there, wakeup time is absolute.
    In the case of Windows, we compute the next sleep period and only sleep if
    needed.  However, due to thread scheduling issues, we still may be "late"
    (but will correct over time).

    \throws      SCXInternalErrorException in case of internal error
*/
    enum SCXCondition::eConditionResult SCXCondition::Wait()
    {
        enum SCXCondition::eConditionResult returnCode = SCXCondition::eCondTestPredicate;

        // Be certain that SetSleep() has already been called
        if (!m_bSetCalled)
        {
            throw SCXInternalErrorException(L"SetSleep() method has not yet been called", SCXSRCLOCATION);
        }

#if defined(SCX_UNIX)
        // We do this check because pthread_cond_timedwait may wakeup prematurely

        if (m_bRecomputeSleep)
        {
            // Adjust wakeup time for the next interval of sleep period
            m_ts.tv_nsec += static_cast<long>((m_sleepTime % 1000) * 1000000);  // += [0 .. 1) sec
            m_ts.tv_sec += static_cast<time_t>((m_sleepTime / 1000) + (m_ts.tv_nsec / 1000000000L));  // += x sec
            m_ts.tv_nsec %= 1000000000L;  // make it [0..1) sec
        }

        int err;

        if (m_sleepTime)
        {
            // If there's a sleepTime, wait until that time passes
            err = pthread_cond_timedwait(&m_cond, &m_lock, &m_ts);
        }
        else
        {
            // No sleep at all?  That's a sleep until signal ...
            err = pthread_cond_wait(&m_cond, &m_lock);
        }
        if (err != 0)
        {
            if (ETIMEDOUT == err)
            {
                // We got a timeout, so recompute next time
                m_bRecomputeSleep = true;

                returnCode = SCXCondition::eCondTimeout;
            }
            else
            {
                throw SCXErrnoException(L"pthread_cond_timedwait() function call failed", err, SCXSRCLOCATION);
            }
        }
#elif defined(WIN32)
        if (m_sleepTime == 0)
        {
            // Unlock our mutex
            Unlock();

            // Wait for the object
            DWORD err = WaitForSingleObject(m_cond, INFINITE);
            DWORD exErr = GetLastError(); // Save extended error information

            // Lock our mutex
            Lock();

            if (err != WAIT_OBJECT_0)
            {
                // We presumably got WAIT_FAILED (not possible to get WAIT_TIMEOUT in this case)
                throw SCXErrnoException(L"WaitForSingleObject() function call failed", exErr, SCXSRCLOCATION);
            }
        }
        else {
            scxulong timestamp = GetMillisecondTimeStamp();

            // Check that the new target time is in the future
            if (m_timeStamp + m_sleepTime > timestamp)
            {
                // If in the future, calculate the time to sleep to reach it
                scxulong sleeptime = m_timeStamp + m_sleepTime - timestamp;

                // Unlock our mutex
                Unlock();

                // Wait for the object
                DWORD err = WaitForSingleObject(m_cond, static_cast<DWORD> (sleeptime));
                DWORD exErr = GetLastError(); // Save extended error information

                // Lock the mutex
                Lock();

                if (err == WAIT_OBJECT_0)
                {
                    // Fall through - we were signaled normally
                }
                else if (err == WAIT_TIMEOUT)
                {
                    returnCode = SCXCondition::eCondTimeout;
                }
                else
                {
                    // We presumably got WAIT_FAILED?
                    throw SCXErrnoException(L"WaitForSingleObject() function call failed", exErr, SCXSRCLOCATION);
                }
            }
            else
            {
                // If not in the future, we should not sleep at all
                // Set caculated after sleep time to use as basis for next sleep

                returnCode = SCXCondition::eCondTimeout;
            }

            m_timeStamp = m_timeStamp + m_sleepTime;
        }
#else
#error Platform not defined
#endif

        return returnCode;
    }

/*----------------------------------------------------------------------------*/
/**
    End a condition test

    \throws      SCXInternalErrorException if we are not in a condition
*/
    void SCXCondition::EndCondition()
    {
        // Unlock our mutex
        Unlock();
    }

/*----------------------------------------------------------------------------*/
/**
    Signal a condition (causes code waiting on a condition to wake up).

    Prior to signaling a condition, the signaler is responsible to set the
    predicate (the shared data value between the signaler and the waiter).

    \throws      SCXErrnoException if error occurred calling system function

*/
    void SCXCondition::Signal()
    {
#if defined(SCX_UNIX)
        int err;
        if (0 != (err = pthread_cond_signal(&m_cond)))
        {
            throw SCXErrnoException(L"pthread_cond_signal() function call failed", err, SCXSRCLOCATION);
        }
#elif defined(WIN32)
        if (0 == PulseEvent(m_cond))
        {
            throw SCXErrnoException(L"PulseEvent() function call failed", GetLastError(), SCXSRCLOCATION);
        }
#else
#error Platform not implemented
#endif
    }

/*----------------------------------------------------------------------------*/
/**
    Get current time in milliseconds

    \returns     A scxulong representing current time in milliseconds
    \throws      SCXInternalErrorException if system functions fail
*/
    scxulong SCXCondition::GetMillisecondTimeStamp()
    {
#if defined(sun) || defined(linux) || defined(hpux) || defined(aix)
        // On UNIX, we should only be called once (m_ts is used in Wait())
        SCXASSERT( 0 == m_ts.tv_sec );
        SCXASSERT( 0 == m_ts.tv_nsec );

        if (0 != clock_gettime(CLOCK_REALTIME, &m_ts))
        {
            throw SCXInternalErrorException(L"clock_gettime() failed", SCXSRCLOCATION);
        }

        return static_cast<scxulong>(m_ts.tv_sec) * 1000 + m_ts.tv_nsec / 1000000;
#elif defined(macos)
        /*
          MacOS (Darwin) doesn't have clock_gettime().

          Gettimeofday() returns a timespec (timeval?) with secs, uSecs (need to convert to nSec?)

          Also, check:

          http://developer.apple.com/documentation/Darwin/Conceptual/KernelProgramming/services/chapter_16_section_2.html

          for clock_get_system_nanotime  that might be useful too.

          Here's some sample code that works on Mac OS/X.  This has NOT been factored into
          this module as Mac OS/X isn't a supported platform yet:

          TIMEVAL_TO_TIMESPEC is a time.h macro.

          int PThreadCondition::timedwait(const struct timeval& time, int& res)
          {
              struct timespec ts;
              struct timeval  tv;
              struct timezone tz;

              gettimeofday(&tv, &tz);
              TIMEVAL_TO_TIMESPEC(&tv, &ts);
              ts.tv_sec += time.tv_sec;

              ts.tv_nsec += time.tv_usec * 1000;
              if (ts.tv_nsec > 1000000000)
              {
                  ts.tv_nsec -= 1000000000;
                  ts.tv_sec += 1;
              }

              pthread_mutex_t& mt = _mutex.mutex();
              res = pthread_cond_timedwait(&_cond, &mt, &ts);
              if (res == ETIMEDOUT)
                  res = WaitResult::timedOut;
              else if (res == EINVAL)
                  res = WaitResult::failed;
              else if (res == 0)
                  res = WaitResult::signalled;
              return res;
          }
         */

        struct timeval tv;

        if (0 != gettimeofday(&tv, NULL))
        {
            throw SCXInternalErrorException(L"gettimeofday() failed", SCXSRCLOCATION);
        }

        return static_cast<scxulong>(tv.tv_sec) * 1000 + tv.tv_usec / 1000;
#elif defined(WIN32)
        return GetTickCount();
#else
#error "Plattform not implemented"
#endif
    }

/**
    Locks a condition.

    This is handled automatically by either the SCXCondition class itself or
    by the SCXConditionHandle helper class.

    \throws      SCXErrnoException if error occurred calling system function
 */
    void SCXCondition::Lock()
    {
#if defined(SCX_UNIX)
        int err;
        if (0 != (err = pthread_mutex_lock(&m_lock)))
        {
            throw SCXErrnoException(L"pthread_mutex_lock() function call failed", err, SCXSRCLOCATION);
        }
#elif defined(WIN32)
        EnterCriticalSection(&m_lock);
#else
#error Platform not supported
#endif
    }

/**
    Unlocks a condition.

    This is handled automatically by either the SCXCondition class itself or
    by the SCXConditionHandle helper class.

    \throws      SCXErrnoException if error occurred calling system function
 */
    void SCXCondition::Unlock()
    {
#if defined(SCX_UNIX)
        int err;
        if (0 != (err = pthread_mutex_unlock(&m_lock)))
        {
            throw SCXErrnoException(L"pthread_mutex_unlock() function call failed", err, SCXSRCLOCATION);
        }
#elif defined(WIN32)
        LeaveCriticalSection(&m_lock);
#else
#error Platform not supported
#endif
        }
}

/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
