/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file        scxprocess.cpp

    \brief       Implements the process handling PAL. 
    
    \date        2007-10-12 15:43:00
 
*/
/*----------------------------------------------------------------------------*/

#if defined(hpux)
#undef _XOPEN_SOURCE // Or else chroot is not defined.
#include <sys/time.h>
#endif
#include <scxcorelib/scxcmn.h>
#include <scxcorelib/scxprocess.h>
#include <scxcorelib/scxoserror.h>
#include <scxcorelib/stringaid.h>
#include <scxcorelib/scxmath.h>
#include <scxcorelib/scxthread.h>
#include <iostream>
#include <vector>
#include <stdio.h>
#if defined(WIN32)
    #include <process.h>
#else
    #include <unistd.h>
    #include <sys/wait.h>
#endif

#if defined(macos) || defined(sun)
#include <signal.h>
#endif

#include <errno.h>

namespace SCXCoreLib
{
 
    /*----------------------------------------------------------------------------*/
    /**
        Retrieve the calling process', process id.
    
        \returns      SCXProcessId of the calling thread.
        
    */
    SCXProcessId SCXProcess::GetCurrentProcessID()
    {
#if defined(WIN32)
        return GetCurrentProcessId();
#elif defined(SCX_UNIX)
        return getpid();
#else
#error "Platform not supported"
#endif
    }
     
    //! Split a command into its separate parts
    //! \param[in]  command like those entered in a command shell
    //! \returns    Parts that was separated by space in "command"
    //! A part is allowed to contain spaces if those are quoted by single or double quotes.
    //! If a part itself consists of a single quote, that part may be quoted by a double quote,
    //! and vice versa.
    //! \note   Parts are delimited by spaces, not the quotes. If one were to list all files
    //          of the directory /usr/local/apache-tomcat/, it could be written as follows:
    //!         ls "/usr/"local/'apache-tomcat/'. That is useful if a command consists of 
    //!         single as well as double quotes.
    std::vector<std::wstring> SCXProcess::SplitCommand(const std::wstring &command) {
        std::vector<std::wstring> parts;
        
        std::wostringstream part;
        bool newPart = false;
        wchar_t quote = 0;
        for (std::wstring::size_type i = 0; i < command.length(); i++) 
        {
            wchar_t c = command.at(i);        
            if (c == ' ') 
            {
                if (quote) 
                {
                    part << c;
                }
                else if (newPart) 
                {
                    parts.push_back(part.str());
                    part.str(L"");                    
                    newPart = false;
                }
            } 
            else if (c == '\'' || c == '"') 
            {
                if (quote == c)   
                {
                    quote = 0;
                }
                else if (quote)
                {
                    part << c;                    
                }
                else 
                {
                    quote = c;
                }
            }
            else 
            {
                part << c;                
                newPart = true;
            }
        }
        if (newPart) {
            parts.push_back(part.str());      
        }
        return parts;
    }
    
#if defined(SCX_UNIX)

        /**********************************************************************************/
    //! Run a process by passing it arguments and streams for stdin, stdout and stderr
    //! \param[in]  command     Corresponding to what would be entered by a user in a command shell
    //! \param[in]  mystdin     stdin for the process
    //! \param[in]  mystdout    stdout for the process
    //! \param[in]  mystderr    stderr for the process
    //! \param[in]  timeout     Accepted number of millieconds to wait for return
    //! \param[in]  cwd         Directory to be set as current working directory for process.
    //! \param[in]  chrootPath  Directory to be used for chrooting process.
    //! \returns exit code of run process.
    //! \throws     SCXInternalErrorException           Failure that could not be prevented
    //! This call will block as long as the process writes to its stdout or stderr
    //! \note   Make sure that "mystdout" and "mystderr" does not block when written to.
    //! 
    //! \date        2007-12-05 15:43:00
    int SCXProcess::Run(const std::wstring &command,
                        std::istream &mystdin,
                        std::ostream &mystdout,
                        std::ostream &mystderr,
                        unsigned timeout,
                        const SCXFilePath& cwd/* = SCXFilePath()*/,
                        const SCXFilePath& chrootPath /* = SCXFilePath() */)
    {
        return Run(SplitCommand(command), mystdin, mystdout,mystderr, timeout, cwd, chrootPath);
    }
    

    /**********************************************************************************/
    //! Class for running a process with an option to timeout 
    class ProcessThreadParam : public SCXThreadParam
    {   
    public:
        //! Constructor that starts a process
        //! \param[in]  process     Process to monitor
        //! \param[in]  timeout     Accepted wait before process terminates
        ProcessThreadParam(SCXProcess *process, unsigned timeout)
            : m_process(process), m_processTerminated(false), m_timeout(timeout)
        {
        }

        /**********************************************************************************/
        //! Wait for the process to terminate
        //! \param[in]  mystdin     Represents stdin of process
        //! \param[in]  mystdout    Represents stdout of process
        //! \param[in]  mystderr    Represents stderr of process
        int WaitForReturn(std::istream &mystdin, std::ostream &mystdout, std::ostream &mystderr)
        {               
            int returnCode = -1;
            try 
            {
                returnCode =  m_process->WaitForReturn(mystdin, mystdout, mystderr);
                m_processTerminated = true;
            }
            catch (...)
            {
                m_processTerminated = true;
                throw;
            }
            return returnCode;
        }

        /**********************************************************************************/
        //! Wait a limited amount of time for the process to terminate.
        void WaitForReturn() 
        {
            // Good enough way of waiting for the process
            int timeoutLeft = m_timeout;             
            while (timeoutLeft > 0 && !m_processTerminated)
            {
                const int timeBetweenChecks = 1000;
                SCXThread::Sleep(timeBetweenChecks);
                timeoutLeft -= timeBetweenChecks;
            }
            if (!m_processTerminated)
            {
                m_process->Kill();
            }
        }

    private:
        ProcessThreadParam(const ProcessThreadParam &);             //!< Prevent copying               
        ProcessThreadParam &operator=(const ProcessThreadParam &);  //!< Prevent assignment

        SCXProcess *m_process;      //!< Process to run
        bool m_processTerminated;   //!< True iff the process has terminated
        unsigned m_timeout;              //!< Accepted wait before termination
    };

    /**********************************************************************************/
    //! Monitors the process
    //! \param[in]  handle      Parameters of thread
    //! \note Meant to be called in a separate thread by beeing passed as argument to Thread constructor
    void WaitForReturnFn(SCXCoreLib::SCXThreadParamHandle& handle)
    {        
        ProcessThreadParam* param = static_cast<ProcessThreadParam *>(handle.GetData());
        param->WaitForReturn();
    }

    /**********************************************************************************/
    //! Run a process by passing it arguments and streams for stdin, stdout and stderr
    //! \param[in]  myargv      Arguments corresponding to argv in the main function of the process
    //! \param[in]  mystdin     stdin for the process
    //! \param[in]  mystdout    stdout for the process
    //! \param[in]  mystderr    stderr for the process
    //! \param[in]  timeout     Max number of milliseconds the process is allowed to run (0 means no limit)
    //! \param[in]  cwd         Directory to be set as current working directory for process.
    //! \param[in]  chrootPath  Directory to be used for chrooting process.
    //! \returns exit code of run process.
    //! \throws     SCXInternalErrorException           Failure that could not be prevented
    //! This call will block as long as the process writes to its stdout or stderr
    //! \note   Make sure that "mystdout" and "mystderr" does not block when written to.
    //! 
    //! \date        2007-12-05 15:43:00
    int SCXProcess::Run(const std::vector<std::wstring> &myargv,
                        std::istream &mystdin, std::ostream &mystdout, std::ostream &mystderr,
                        unsigned timeout,
                        const SCXFilePath& cwd/* = SCXFilePath()*/,
                        const SCXFilePath& chrootPath /* = SCXFilePath() */) 
    {
        int returnCode = -1;        
        SCXProcess process(myargv, cwd, chrootPath);
        if (timeout <= 0) 
        {
            returnCode = process.WaitForReturn(mystdin, mystdout, mystderr);
        }
        else
        {           
            ProcessThreadParam *ptp = new ProcessThreadParam(&process, timeout);
            //ptp = 0;
            SCXThreadParamHandle tph(ptp);
            
            SCXThread t(WaitForReturnFn, tph);        
            returnCode = ptp->WaitForReturn(mystdin, mystdout, mystderr);
            t.Wait();
            
        } 
        return returnCode;
    }

    /**********************************************************************************/
    //! Constructor
    //! \param[in]  myargv       Arguments to process
    //! \param[in]  cwd         Directory to be set as current working directory for process.
    //! \param[in]  chrootPath  Directory to be used for chrooting process.

    SCXProcess::SCXProcess(const std::vector<std::wstring> myargv,
                           const SCXFilePath& cwd/* = SCXFilePath()*/,
                           const SCXFilePath& chrootPath /* = SCXFilePath() */) :
            m_stdinChars(1000),
            m_buffer(1000),
            m_stdinCharCount(0),
            m_pid(-1)
        {

        // Convert arguments to types expected by the system function for running processes
        for (std::vector<char *>::size_type i = 0; i < myargv.size(); i++) 
        {
            m_cargv.push_back(strdup(StrToMultibyte(myargv[i]).c_str()));
        }
        m_cargv.push_back(0);
        
        // Create pipes for communicating with the child process
        pipe(m_inForChild);
        pipe(m_outForChild);        
        pipe(m_errForChild);         
        m_pid = fork();                         // Create child process, duplicates file descriptors       
        if (m_pid == 0) 
        {
            // Child process. 
            // The file descriptors are duplicates, created by "fork",  of those in the parent process
            dup2(m_inForChild[R], STDIN_FILENO);      // Make the child process read from the parent process
            close(m_inForChild[R]);                   // Close duplicate
            close(m_inForChild[W]);                   // The child only reads from its stdin
            dup2(m_outForChild[W], STDOUT_FILENO);    // Make the child process write to the parent process
            close(m_outForChild[R]);                  // The child only writes to its stdout
            close(m_outForChild[W]);                  // Close duplicate
            dup2(m_errForChild[W], STDERR_FILENO);    // Make the child process write to the parent process
            close(m_errForChild[R]);                  // The child only writes to its stdout
            close(m_errForChild[W]);                  // Close duplicate

            char error_msg[1024];
            if (L"" != chrootPath.Get())
            {
                if ( 0 != ::chroot(StrToMultibyte(chrootPath.Get()).c_str()))
                {
                    snprintf(error_msg,
                             sizeof(error_msg),
                             "Failed to chroot '%s' errno=%d",
                             StrToMultibyte(chrootPath.Get()).c_str(), errno);
                    write(STDERR_FILENO, error_msg, strlen(error_msg));
                    exit(1);
                }
                if ( 0 != ::chdir("/"))
                {
                    snprintf(error_msg,
                             sizeof(error_msg),
                             "Failed to change root directory. errno=%d", errno);
                    write(STDERR_FILENO, error_msg, strlen(error_msg));
                    exit(1);
                }
            }
            if (L"" != cwd.Get())
            {
                if (0 != ::chdir(StrToMultibyte(cwd.Get()).c_str()))
                {
                    snprintf(error_msg,
                             sizeof(error_msg),
                             "Failed to change cwd. errno=%d", errno);
                    write(STDERR_FILENO, error_msg, strlen(error_msg));
                    exit(1);
                }
            }

            // Close open file descriptors except stdin/out/err
            // (Some systems have UNLIMITED of 2^64; limit to something reasonable)

            int fdLimit = 0;
            fdLimit = getdtablesize();
            if (fdLimit > 2500)
            {
                fdLimit = 2500;
            }

            for (int fd = 3; fd < fdLimit; ++fd)
            {
                close(fd);
            }

            execvp(m_cargv[0], &m_cargv[0]);                // Replace the child process image
            snprintf(error_msg, sizeof(error_msg), "Failed to start child process '%s' errno=%d", m_cargv[0], errno);
            write(STDERR_FILENO, error_msg, strlen(error_msg));
            exit(1);                                // Failed to load correct process.
        } 
        else 
        {
            // Same (parent) process.  
            // All file descriptors were duplicated in the child process by "fork"
            close(m_inForChild[R]);                   // Only the child process reads from its stdin   
            close(m_outForChild[W]);                  // Only the child process writes to its stdout
            close(m_errForChild[W]);                  // Only the child process writes to its stderr
            if (m_pid < 0) 
            {
                // Free remaining resources held by the parent process
                // The child process manages its own resources
                close(m_inForChild[W]);
                close(m_outForChild[R]);
                close(m_errForChild[R]);

                // Free everything except the terminating NULL
                for (std::vector<char *>::size_type i = 0; i < m_cargv.size() - 1; i++) 
                {
                    free(m_cargv[i]);
                }
                //  No child process was created
                throw SCXInternalErrorException(UnexpectedErrno(L"Process communication failed", errno), SCXSRCLOCATION);

            }                         
        }
        }
    
    /**********************************************************************************/
    //! Make the process fetch input
    //! \param[in]  mystdin     Source
    void SCXProcess::FetchInput(std::istream &mystdin) {
        std::streamsize stdinSomeCharCount = mystdin.readsome(&m_stdinChars[m_stdinCharCount], m_stdinChars.size() - m_stdinCharCount);
        if (!mystdin.eof() && !mystdin.good()) 
        {
            throw SCXInternalErrorException(L"Process parent communication failed", SCXSRCLOCATION);            
        }
        m_stdinCharCount += static_cast<int>(stdinSomeCharCount);
        ssize_t bytesWritten = write(m_inForChild[W], &m_stdinChars[0], m_stdinCharCount);
        if (bytesWritten < 0) 
        {
            throw SCXInternalErrorException(UnexpectedErrno(L"Process communication failed", errno), SCXSRCLOCATION);
        }
        strncpy(&m_stdinChars[0], &m_stdinChars[bytesWritten], m_stdinCharCount - bytesWritten);
        m_stdinCharCount -= static_cast<int>(bytesWritten);
    }
    
    /**********************************************************************************/
    //! Fetch a block of output from process
    //! \param[in]  mystdout    Receiver of content that the process writes to stdout
    //! \param[in]  mystderr    Receiver of content that the process writes to stderr
    bool SCXProcess::FetchOutput(std::ostream &mystdout, std::ostream &mystderr) {
        const int maxSecondsBetweenStdinReads = 2;     // Maximum contigous pause when waiting for the child process to write to stdout or stderr   
        // Wait some time for something available to be read from stdout or stderr of the child process
        bool stdoutActive = true;               // The child process may write to its stdout
        bool stderrActive = true;               // The child process may write to its stderr
        fd_set readfds;
        FD_ZERO(&readfds);
        FD_SET(m_outForChild[R], &readfds);
        FD_SET(m_errForChild[R], &readfds);
        int maxFd = MAX(m_outForChild[R], m_errForChild[R]);
        timeval timeout;
        timeout.tv_sec = maxSecondsBetweenStdinReads;
        timeout.tv_usec = 0;           
        if (select(maxFd + 1, &readfds, NULL, NULL, &timeout) < 0) 
        {
            throw SCXInternalErrorException(UnexpectedErrno(L"Process communication failed", errno), SCXSRCLOCATION);
        }                
        
        // Move data from the child process stdout to the stdout stream parameter
        if (FD_ISSET(m_outForChild[R], &readfds)) 
        {
            ssize_t bytesRead = read(m_outForChild[R], &m_buffer[0], m_buffer.size());
            if (bytesRead > 0) {
                mystdout.write(&m_buffer[0], bytesRead);
            } else if (bytesRead < 0) {
                throw SCXInternalErrorException(UnexpectedErrno(L"Process communication failed", errno), SCXSRCLOCATION);
            } else {
                stdoutActive = false;                
            }
        }
       
        // Move data from the child process stderr to the stdout stream parameter
        if (FD_ISSET(m_errForChild[R], &readfds)) 
        {
            ssize_t bytesRead = read(m_errForChild[R], &m_buffer[0], m_buffer.size());
            if (bytesRead > 0) {
                mystderr.write(&m_buffer[0], bytesRead);
            } else if (bytesRead < 0) {
                throw SCXInternalErrorException(UnexpectedErrno(L"Process communication failed", errno), SCXSRCLOCATION);
            } else {
                stderrActive = false;                
            }
        }
        return stdoutActive || stderrActive;
    }
    
    /**********************************************************************************/
    //! Wait for the process to return, that is, terminate normally or by beeing signaled
    //! ±throws     SCXInternalErrorException       Did not succeed to wait for the process
    int SCXProcess::WaitForReturn() 
    {
        int child_status = 0;
        if (m_pid != waitpid(m_pid, &child_status, 0))
        {
            throw SCXInternalErrorException(UnexpectedErrno(L"Failed to wait for child process", errno), SCXSRCLOCATION);
        }
        if (!WIFEXITED(child_status)) 
        {
            throw SCXInterruptedProcessException(SCXSRCLOCATION);
        }
        return WEXITSTATUS(child_status);       
    }
    
    /**********************************************************************************/
    //! Interacts with the process while waiting for the it to return, 
    //! that is, terminate normally or by beeing signaled
    //! \param[in]  mystdin     Stdin of process
    //! \param[in]  mystdout    Stdout of process
    //! \param[in]  mystderr    Stderr of process
    int SCXProcess::WaitForReturn(std::istream &mystdin, std::ostream &mystdout, std::ostream &mystderr)
    {
        FetchInput(mystdin);
        bool fetched = FetchOutput(mystdout, mystderr);
        while (fetched) 
        {
            FetchInput(mystdin);            
            fetched = FetchOutput(mystdout, mystderr);                    
        }           
        return WaitForReturn();
        
    }
    
    /**********************************************************************************/
    //! Destructor
    SCXProcess::~SCXProcess() 
    {
        // Free remaining resources held by the parent process
        // The child process manages its own resources
        close(m_inForChild[W]);
        close(m_outForChild[R]);
        close(m_errForChild[R]);

        // Free everything except the terminating NULL
        for (std::vector<char *>::size_type i = 0; i < m_cargv.size() - 1; i++) 
        {
            free(m_cargv[i]);
        }
    }

    /**********************************************************************************/
    //! Terminate the process
    void SCXProcess::Kill() 
    {
        if (kill(m_pid, SIGKILL) < 0 && errno != ESRCH) 
        {
                        
            throw SCXInternalErrorException(UnexpectedErrno(L"Unable to kill child process", errno), SCXSRCLOCATION);
        }
    }
    
#endif


} /* namespace SCXCoreLib */

/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
