/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file        

    \brief         Implements the thread lock factory PAL.

    \date          2007-05-28 10:46:00
  
*/
/*----------------------------------------------------------------------*/
#include <scxcorelib/scxcmn.h>
#include <scxcorelib/scxthreadlock.h>
#include <scxcorelib/stringaid.h>

namespace SCXCoreLib
{
    SCXThreadLockFactory SCXThreadLockFactory::s_instance;

/*----------------------------------------------------------------------------*/
/**
    Convenience function to access the thread lock factory and get an anonymous
    lock handle.
    
    Parameters:  None
    Retval:      An anonymous SCXThreadLockHandle.
        
*/
    SCXThreadLockHandle ThreadLockHandleGet(void)
    {
        return SCXThreadLockFactory::GetInstance().GetLock();
    }

/*----------------------------------------------------------------------------*/
/**
    Convenience function to access the thread lock factory and get a named
    lock handle.
    
    Parameters:  nameOfLock - name of lock handle to get.
    Retval:      A SCXThreadLockHandle associated with the given name.
        
    Note that calling this method with an empty string generates the same result
    as calling ThreadLockFactory(void).
    
*/
    SCXThreadLockHandle ThreadLockHandleGet(const std::wstring& nameOfLock)
    {
        return SCXThreadLockFactory::GetInstance().GetLock(nameOfLock);
    }

/*----------------------------------------------------------------------------*/
/**
    Default constructor.
    
    Parameters:  None
    Retval:      N/A
        
    Creates a single anonymous lock handle to be used as thread lock to make the
    factory class thread safe.
    
*/
    SCXThreadLockFactory::SCXThreadLockFactory(void):
        m_lockHandle(L"")
    {
        Reset();
    }

/*----------------------------------------------------------------------------*/
/**
    Virtual destructor.
    
    Parameters:  None
    Retval:      N/A
        
    Releases all allocated resources.
    
*/
    SCXThreadLockFactory::~SCXThreadLockFactory(void)
    {
        Reset();
    }

/*----------------------------------------------------------------------------*/
/**
    Dump object as string (for logging).
    
    Parameters:  None
    Retval:      N/A
        
*/
    const std::wstring SCXThreadLockFactory::DumpString() const
    {
        return L"SCXThreadLockFactory locks=" + SCXCoreLib::StrFrom(static_cast<scxlong>(m_locks.size()));
    }

/*----------------------------------------------------------------------------*/
/**
    Static method to retrieve singleton instance.
    
    Parameters:  None
    Retval:      Singleton instance of factory class.
        
*/
    SCXThreadLockFactory& SCXThreadLockFactory::GetInstance(void)
    {
        return s_instance;
    }

/*----------------------------------------------------------------------------*/
/**
    Create an anonymous thread lock handle.
    
    Parameters:  None
    Retval:      An anonymous SCXThreadLockHandle object.
        
*/
    SCXThreadLockHandle SCXThreadLockFactory::GetLock(void)
    {
        SCXThreadLockHandle l(L"");
        return l;
    }

/*----------------------------------------------------------------------------*/
/**
    Retrieve a named thread lock handle.
    
    Parameters:  nameOfLock - Name of lock to retrieve.
    Retval:      A SCXThreadLockHandle associated with the given name.
        
    Using this method with an empty lock name will create an anonymous lock handles.
    If name is non-empty the internal list is searched for the given name. The
    found lock handle is returned if one is found. Otherwise a new lock handle
    is created and added to the list.
    
*/
    SCXThreadLockHandle SCXThreadLockFactory::GetLock(const std::wstring& nameOfLock)
    {
        if (nameOfLock.empty())
            return GetLock();

        SCXThreadLock lock(m_lockHandle);

        const std::map<std::wstring,SCXThreadLockHandle>::iterator item = m_locks.find(nameOfLock);

        if (item != m_locks.end())
        {
            return (item->second);
        }
        
        SCXThreadLockHandle l(nameOfLock);
        m_locks[nameOfLock] = l;
        return l;
    }

/*----------------------------------------------------------------------------*/
/**
    Reset the factory.
    
    Parameters:  None
    Retval:      None
        
    Will remove all references to any previously created locks. In practice this
    method is probably only usable to ease memory leak detection.
    
*/
    void SCXThreadLockFactory::Reset(void)
    {
        SCXThreadLock lock(m_lockHandle);

        m_locks.clear();
    }

/*----------------------------------------------------------------------------*/
/**
    Get the number of used locks.
    
    Parameters:  None
    Retval:      Number of loch handles in used.
        
    Will traverse the lock handle list in order to count number of used locks.
    A lock is considered to be in use if its reference counter is not one since
    one reference is that in the factory lock handle list.
    
*/
    unsigned int SCXThreadLockFactory::GetLocksUsed(void) const
    {
        SCXThreadLock lock(m_lockHandle);

        unsigned int r = 0;
        std::map<std::wstring,SCXThreadLockHandle>::const_iterator cur = m_locks.begin();
        std::map<std::wstring,SCXThreadLockHandle>::const_iterator end = m_locks.end();

        for ( ; cur != end; ++cur)
        {
            if (cur->second.GetRefCount() > 1)
            { /* Ref count should be one for unused locks since they are in the list (=one reference) */
                ++r;
            }
        }
        return r;
    }

} /* namespace SCXCoreLib */
