/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file        

    \brief       Implement a lvmtab file parser.
    
    \date        2008-02-21 13:31:02

    The file format used in lvmtab files is (reverse engineered):
    7 bytes
    1 byte = number of volume groups
    4 bytes
    For each volume group:
        1024 bytes = name of volume group
        17 bytes
        1 byte = number of parts in volume group
        12 bytes
        For each part in volume group:
            1024 bytes = name of part
            4 bytes
    
*/
/*----------------------------------------------------------------------------*/

#include <scxcorelib/scxcmn.h>
#include <scxsystemlib/scxlvmtab.h>
#include <scxcorelib/scxfile.h>
#include <scxcorelib/stringaid.h>

namespace SCXSystemLib
{
/*----------------------------------------------------------------------------*/
/**
    Constructor parsing a given file as an lvmtab file.
    
    \param       path A file path to an lvmtab file.
    \throws      SCXLvmTabFormatException if the file parsed has the wrong format.
    \throws      SCXFilePathNotFoundException if the path could not be opened.
    \throws      SCXUnauthorizedFileSystemAccessException if the file may not be opened.
    
    \note The parser does not fail for files with wrong format. It does a best
    effort try to parse the file.
    
*/
    SCXLvmTab::SCXLvmTab(const SCXCoreLib::SCXFilePath& path)
    {
        SCXCoreLib::SCXHandle<std::fstream> fs = SCXCoreLib::SCXFile::OpenFstream(path, std::ios_base::in);
        fs.SetOwner();
        std::fstream::pos_type file_pos = 0; // HPUX aCC forces us to keep track of this manually to be able to seekg
        // Skip 7 first bytes
        file_pos += 7;
        // Read number of VGs.
        fs->seekg(file_pos);
        char num_vg = 0;
        fs->read(&num_vg, sizeof(num_vg));
        file_pos += sizeof(num_vg);
        // Skip 4 bytes.
        file_pos += 4;
        for (char vg_idx=0; vg_idx < num_vg && SCXCoreLib::SCXStream::IsGood(*fs); ++vg_idx)
        {
            SCXVG vg;
            // Read volume group (1K)
            char volumeGroupRaw[1024+1];
            memset(volumeGroupRaw, 0, sizeof(volumeGroupRaw));
            fs->seekg(file_pos);
            fs->read(volumeGroupRaw, sizeof(volumeGroupRaw)-1);
            file_pos += sizeof(volumeGroupRaw)-1;
            vg.m_name = SCXCoreLib::StrFromMultibyte(std::string(volumeGroupRaw));
            // Skip 17 bytes
            file_pos += 17;
            // Read number of parts
            char num_parts = 0;
            fs->seekg(file_pos);
            fs->read(&num_parts, sizeof(num_parts));
            file_pos += sizeof(num_parts);
            // Skip 12 bytes
            file_pos += 12;
            for (char part_idx = 0; part_idx < num_parts && SCXCoreLib::SCXStream::IsGood(*fs); ++part_idx)
            {
                // Read partition (1K)
                char partitionRaw[1024+1];
                memset(partitionRaw, 0, sizeof(partitionRaw));
                fs->seekg(file_pos);
                fs->read(partitionRaw, sizeof(partitionRaw)-1);
                file_pos += sizeof(partitionRaw)-1;
                vg.m_part.push_back(SCXCoreLib::StrFromMultibyte(std::string(partitionRaw)));
                // Skip 4 bytes.
                file_pos += 4;
            }
            m_vg.push_back(vg);
        }
        
        fs->seekg(0, std::ios_base::end);
        if (file_pos > fs->tellg())
        {
            throw SCXLvmTabFormatException(L"File to short", SCXSRCLOCATION);
        }
        if (file_pos < fs->tellg())
        {
            throw SCXLvmTabFormatException(L"File to long", SCXSRCLOCATION);
        }
        fs->close();
    }

/*----------------------------------------------------------------------------*/
/**
    Return the number of volume groups in the parsed file.
    
    \returns     Number of volume groups.
    
*/
    size_t SCXLvmTab::GetVGCount() const
    {
        return m_vg.size();
    }

/*----------------------------------------------------------------------------*/
/**
    Return the number of partitions in a particular volume group.
    
    \param       vg_idx Volume group index (zero being the first VG in the lvmtab file).
    \returns     Number of partitions in the given volume group.
    \throws      SCXIllegalIndexExceptionUInt if the volume group index is out of bounds.
    
*/
    size_t SCXLvmTab::GetPartCount(size_t vg_idx) const
    {
        if (vg_idx >= m_vg.size())
        {
            throw SCXCoreLib::SCXIllegalIndexException<size_t>(L"vg_idx", vg_idx, 0, true, m_vg.size()-1, true, SCXSRCLOCATION);
        }
        return m_vg[vg_idx].m_part.size();
    }

/*----------------------------------------------------------------------------*/
/**
    Return the name of a volume group.
    
    \param       vg_idx Volume group index (zero being the first VG in the lvmtab file).
    \returns     Name of the given volume group.
    \throws      SCXIllegalIndexExceptionUInt if the volume group index is out of bounds.
    
*/
    const std::wstring& SCXLvmTab::GetVG(size_t vg_idx) const
    {
        if (vg_idx >= m_vg.size())
        {
            throw SCXCoreLib::SCXIllegalIndexException<size_t>(L"vg_idx", vg_idx, 0, true, m_vg.size()-1, true, SCXSRCLOCATION);
        }
        return m_vg[vg_idx].m_name;
    }

/*----------------------------------------------------------------------------*/
/**
    Return the partition name of a given partition in a given volume group.
    
    \param       vg_idx Volume group index (zero being the first VG in the lvmtab file).
    \param       part_idx Partition index (zero being the first partition for the given VG).
    \returns     Name of given partition in given volume group.
    \throws      SCXIllegalIndexExceptionUInt if the volume group or partition index is out of bounds.
    
*/
    const std::wstring& SCXLvmTab::GetPart(size_t vg_idx, size_t part_idx) const
    {
        if (vg_idx >= m_vg.size())
        {
            throw SCXCoreLib::SCXIllegalIndexException<size_t>(L"vg_idx", vg_idx, 0, true, m_vg.size()-1, true, SCXSRCLOCATION);
        }
        if (part_idx >= m_vg[vg_idx].m_part.size())
        {
            throw SCXCoreLib::SCXIllegalIndexException<size_t>(L"part_idx", part_idx, 0, true, m_vg[vg_idx].m_part.size()-1, true, SCXSRCLOCATION);
        }
        return m_vg[vg_idx].m_part[part_idx];
    }

} /* namespace SCXSystemLib */

/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
