/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
   \file        

   \brief       Implements the volume group configuration parser.
    
   \date        2008-01-17 15:39:02

*/
/*----------------------------------------------------------------------------*/

#include <fstream>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <scxcorelib/scxcmn.h>
#include <scxsystemlib/scxvgcfg.h>
#include <scxcorelib/scxstream.h>
#include <scxcorelib/scxfile.h>
#include <scxcorelib/stringaid.h>
#include <scxcorelib/scxdirectoryinfo.h>
#include <scxcorelib/scxthreadlock.h>

namespace SCXSystemLib {

/*----------------------------------------------------------------------------*/
/**
   Default constructor
*/
    SCXVolumeGroupConfigurationItem::SCXVolumeGroupConfigurationItem()
        : m_name(L"UNKNOWN")
    {

    }

    /*----------------------------------------------------------------------------*/
/**
   Virtual destructor.
*/
    SCXVolumeGroupConfigurationItem::~SCXVolumeGroupConfigurationItem()
    {
    }

    /*----------------------------------------------------------------------------*/
/**
   Get name of item.

   \returns     Name of item.
*/
    const std::wstring& SCXVolumeGroupConfigurationItem::GetName() const
    {
        return m_name;
    }

    /*----------------------------------------------------------------------------*/
/**
   Get child items.

   \returns     Vector of child items.
*/
    const SCXVolumeGroupConfigurationItemVector& SCXVolumeGroupConfigurationItem::GetSections() const 
    { 
        return m_sections; 
    }

    /*----------------------------------------------------------------------------*/
/**
   Get attributes.

   \returns     A string to string map with attribute names and values.
*/
    const SCXVolumeGroupConfigurationItemAttributes& SCXVolumeGroupConfigurationItem::GetAttributes() const 
    { 
        return m_attributes; 
    }

    /*----------------------------------------------------------------------------*/
/**
   Parse a configuration file.

   \param       lines - The configuration as a number of lines in a vector. Parsed
                lines are removed from vector.
   \throws      SCXVolumeGroupConfigurationException on parse errors.

   This method consumes lines of the vector in order to parse an item. It adds attributes
   and child items as found. Parsing ends when a section end delimiter (i.e. "}") is found.
   If the end delimiter is not found when there are no more lines, an parse error 
   exception will be thrown.
*/
    void SCXVolumeGroupConfigurationItem::Parse(std::vector<std::wstring>& lines)
    {
        bool partialAttribute = false;
        while ( ! lines.empty())
        {
            std::wstring line = SCXCoreLib::StrTrim(lines.front());
            if (line.find(L"#") != std::wstring::npos)
            { // Removing comment from line.
                line = SCXCoreLib::StrTrim(line.substr(0, line.find(L"#")));
            }
            if (line.length() == 0)
            { // Empty line
                lines.erase(lines.begin());
                continue;
            }
            if (line == L"}")
            { // Found end of section
                if (partialAttribute)
                {
                    throw SCXVolumeGroupConfigurationException(L"End of section in partial attribute", SCXSRCLOCATION);
                }
                lines.erase(lines.begin());
                return;
            }
            
            if (line.find(L"{") != std::wstring::npos)
            {
                if (partialAttribute)
                {
                    throw SCXVolumeGroupConfigurationException(L"New section in partial attribute", SCXSRCLOCATION);
                }
            
                SCXCoreLib::SCXHandle<SCXVolumeGroupConfigurationItem> item( new SCXVolumeGroupConfigurationItem() );
                item->m_name = SCXCoreLib::StrTrim(line.substr(0,line.find(L"{")));
                lines.erase(lines.begin());
                item->Parse(lines);
                m_sections.push_back(item);
                continue;
            }
            partialAttribute =  ! ParseAttribute(line, partialAttribute);
            lines.erase(lines.begin());
        }
        throw SCXVolumeGroupConfigurationException(L"End of section not found", SCXSRCLOCATION);
    }

    /*----------------------------------------------------------------------------*/
/**
   Parse a single attribute.

   \param       line - line to parse.
   \param       partial - indicate if we have a partial attribute read in progress.
                When parsing lines this parameter should typically be the inverse of
                the return value for the previous call to this method.
   \returns     true if attribute was parsed comlete or false for partial parse.
   \throws      SCXVolumeGroupConfigurationException for parse errors.

   Partial attributes are attributes spanning over several lines.
*/
    bool SCXVolumeGroupConfigurationItem::ParseAttribute(std::wstring& line, bool partial)
    {
        if (partial)
        {
            m_attributes[m_partialAttributeName] = m_attributes[m_partialAttributeName] + line;
            return (line == L"]");
        }

        std::vector<std::wstring> parts;
        SCXCoreLib::StrTokenize( line, parts, L"=" );
        if (parts.size() != 2)
        {
            throw SCXVolumeGroupConfigurationException(line, SCXSRCLOCATION);
        }
        m_attributes[parts[0]] = parts[1];
        if (parts[1].substr(0,1) == L"[")
        {
            m_partialAttributeName = parts[0];
            return (parts[1].substr(parts[1].length()-1) == L"]");
        }
        return true;
    }

    /*----------------------------------------------------------------------------*/
/**
   Constructor

   \param       stream - A stream with the configuration data.
   \throws      SCXVolumeGroupConfigurationException on parse errors.

   This constructor will try to parse the stream as configuration data.
   \note This parser assumes the complete configuration fits into the allocated buffer
   currently with the size 4096 bytes.
*/
    SCXVolumeGroupConfiguration::SCXVolumeGroupConfiguration(std::istream& stream)
    {
        m_root.m_name = L"ROOT";

        char data[4096]; 
        memset(data, 0, sizeof(data));
        stream.read(data, sizeof(data)-1);
        std::istringstream data_stream(data);
        std::vector<std::wstring> lines;
        SCXCoreLib::SCXStream::NLFs nlf;
        SCXCoreLib::SCXStream::ReadAllLinesAsUTF8(data_stream, lines, nlf);
        // fake end of section since root section is not really a real section.
        lines.push_back(L"}");
        m_root.Parse(lines);
    }

    /*----------------------------------------------------------------------------*/
/**
   Virtual destructor.
*/
    SCXVolumeGroupConfiguration::~SCXVolumeGroupConfiguration()
    {
    
    }

    /*----------------------------------------------------------------------------*/
/**
   Return the sequence number of the configuration.

   \returns     Sequence number of configuration.
   \throws      SCXVolumeGroupConfigurationException if the configuration does not
                contain a sequence number.
*/
    size_t SCXVolumeGroupConfiguration::GetSeqNo() const
    {
        if (m_root.GetSections().size() != 1)
        {
            throw SCXVolumeGroupConfigurationException(L"No volume group found", SCXSRCLOCATION);
        }
        SCXCoreLib::SCXHandle<SCXVolumeGroupConfigurationItem> item = *(m_root.GetSections().begin());
        if (item->GetAttributes().find(L"seqno") == item->GetAttributes().end())
        {
            throw SCXVolumeGroupConfigurationException(L"No attribute 'seqno' found", SCXSRCLOCATION);
        }
        return SCXCoreLib::StrToUInt(item->GetAttributes().find(L"seqno")->second);
    }

    /*----------------------------------------------------------------------------*/
/**
   Translates a partition to a volume group and logical volume. 

   \param       partition - Partition to translate.
   \param[out]  volumeGroup - volume group partition belongs to.
   \param[out]  logicalVolume - logical volume partition belongs to.
   \throws      SCXVolumeGroupConfigurationException if failing to translate partition.

   Example partition is "/dev/hda2" and example output is "VolGroup00" and "LogVol00".
*/
    void SCXVolumeGroupConfiguration::MapPartition(const std::wstring& partition, std::wstring& volumeGroup, std::wstring& logicalVolume) const
    {
        if (m_root.GetSections().size() != 1)
        {
            throw SCXVolumeGroupConfigurationException(L"No volume group found", SCXSRCLOCATION);
        }
        SCXCoreLib::SCXHandle<SCXVolumeGroupConfigurationItem> item = *(m_root.GetSections().begin());
        volumeGroup = item->GetName();
        // Find the physical and logical volumes.
        SCXCoreLib::SCXHandle<SCXVolumeGroupConfigurationItem> pvs(0);
        SCXCoreLib::SCXHandle<SCXVolumeGroupConfigurationItem> lvs(0);
        for (SCXVolumeGroupConfigurationItemVector::const_iterator it = item->GetSections().begin();
             it != item->GetSections().end();
             ++it)
        {
            if (0 == SCXCoreLib::StrCompare((*it)->GetName(), L"physical_volumes"))
            {
                pvs = *it;
            }
            else if (0 == SCXCoreLib::StrCompare((*it)->GetName(), L"logical_volumes"))
            {
                lvs = *it;
            }
        }
        if ((0 == pvs) || (0 == lvs))
        {
            throw SCXVolumeGroupConfigurationException(L"No volumes found", SCXSRCLOCATION);
        }
        // Search for partition name among physical volumes.
        std::wstring pv = L"";
        for (SCXVolumeGroupConfigurationItemVector::const_iterator it = pvs->GetSections().begin();
             it != pvs->GetSections().end() && pv.length() == 0;
             ++it)
        {
            if ((*it)->GetAttributes().find(L"device") != (*it)->GetAttributes().end())
            {
                if (0 == SCXCoreLib::StrCompare(SCXCoreLib::StrStrip((*it)->GetAttributes().find(L"device")->second, L"\""), partition))
                {
                    pv = (*it)->GetName();
                }
            }
        }
        if (pv.length() == 0)
        {
            throw SCXVolumeGroupConfigurationException(L"No physical volume found", SCXSRCLOCATION);
        }

        // Search all logical volumes for the physical volume found.
        logicalVolume = L"";
        for (SCXVolumeGroupConfigurationItemVector::const_iterator it = lvs->GetSections().begin();
             it != lvs->GetSections().end() && logicalVolume.length() == 0;
             ++it)
        {
            // Search the segments of the logical volume for the physical volume
            SCXCoreLib::SCXHandle<SCXVolumeGroupConfigurationItem> lv = *it;
            for (SCXVolumeGroupConfigurationItemVector::const_iterator it2 = lv->GetSections().begin();
                 it2 != lv->GetSections().end() && logicalVolume.length() == 0;
                 ++it2)
            {
                if ((*it2)->GetAttributes().find(L"stripes") != (*it2)->GetAttributes().end())
                {
                    std::wstringstream str;
                    str << L"\"" << pv << "\"";
                    if ((*it2)->GetAttributes().find(L"stripes")->second.find(str.str()) != std::wstring::npos)
                    {
                        logicalVolume = lv->GetName();
                    }
                }
            }
        }
        if (logicalVolume.length() == 0)
        {
            throw SCXVolumeGroupConfigurationException(L"No logical volume found", SCXSRCLOCATION);
        }
    }

    /*----------------------------------------------------------------------------*/
/**
   Find partitions which is part of given volume group and logical volume.

   \param[in]   volumeGroup - Volume group.
   \param[in]   logicalVolume - Logical volume.
   \returns     partition paths found.
*/
    std::vector<std::wstring> SCXVolumeGroupConfiguration::FindPartitions(const std::wstring& volumeGroup, const std::wstring& logicalVolume)
    {
        static std::map<unsigned int, std::wstring> s_device_mapper;

        if (s_device_mapper.size() == 0)
        { // Probably the first time - parse /proc/devices
            SCXCoreLib::SCXThreadLock lock(SCXCoreLib::ThreadLockHandleGet(L"SCXVolumeGroupConfiguration::FindPartitions"));
            SCXCoreLib::SCXStream::NLFs nlfs;
            std::vector<std::wstring> lines;
            SCXCoreLib::SCXFile::ReadAllLines(L"/proc/devices", lines, nlfs);
            for (std::vector<std::wstring>::const_iterator it = lines.begin(); it != lines.end(); it++)
            {
                try
                {
                    unsigned int dev = SCXCoreLib::StrToUInt(SCXCoreLib::StrTrim(it->substr(0,4)));
                    std::wstring desc = SCXCoreLib::StrTrim(it->substr(4));
                    if (desc.substr(0,3) == L"ide" || desc == L"sd")
                    { // Only add SCSI & IDE devices IDs to device mapper.
                        s_device_mapper[dev] = desc;
                    }
                }
                catch (SCXCoreLib::SCXNotSupportedException&)
                {
                    continue;
                }
            } 
        }

        std::vector<std::wstring> result;
        std::vector<SCXCoreLib::SCXFilePath> files;
        try
        {
            files = SCXCoreLib::SCXDirectory::GetFileSystemEntries(L"/dev/", SCXCoreLib::eDirSearchOptionFile | SCXCoreLib::eDirSearchOptionSys);
        }
        catch (SCXCoreLib::SCXFilePathNotFoundException&)
        {
            throw SCXVolumeGroupConfigurationException(L"Unable to scan /dev", SCXSRCLOCATION);
        }
        for (size_t i=0; i<files.size(); ++i)
        {
#if defined(PF_DISTRO_REDHAT) && (PF_MAJOR==4)
            // Since RHEL4 kernel logs when we try to open a cd device sometimes (WI 12716)
            // we must do something. And we actually only want partitions (not disk devices)
            // so this check could be generic but is not since we're close to lockdown for RTM.
            if ( ! IsPartition(files[i].Get()))
            {
                continue;
            }
#endif

            struct stat s;
            memset(&s, 0, sizeof(s));
            if (-1 == stat(SCXCoreLib::SCXFileSystem::EncodePath(files[i].Get()).c_str(), &s)) 
            {
                throw SCXCoreLib::SCXErrnoException(L"stat(" + files[i].Get() + L")", errno, SCXSRCLOCATION);
            }
            
            // Shifting rdev value 8 bits to get only major device number
            if (s_device_mapper.find(static_cast<unsigned int>(s.st_rdev>>8)) == s_device_mapper.end())
            { // Device id should not be checked
                continue;
            }

            char buf[255*512];
            try
            {
                if (sizeof(buf) != SCXCoreLib::SCXFile::ReadAvailableBytes(files[i], buf, sizeof(buf)))
                { // If not able to read all 255 512 byte blocks it cannot be a volume group device.
                    continue;
                }
            }
            catch (SCXCoreLib::SCXErrnoException& e)
            { //See WI 12787 for improvements.
                if (e.ErrorNumber() == EACCES /* No permisions */ || 
                    e.ErrorNumber() == ENXIO /* device cannot be opened non-blocking*/ ||
                    e.ErrorNumber() == EIO /* Some IO error */ ||
                    e.ErrorNumber() == EBUSY /* Device busy */ ||
#if defined(linux)
                    e.ErrorNumber() == ENOMEDIUM /* No medium available, i.e. no cd in cd tray */ ||
#endif
                    e.ErrorNumber() == EINVAL /* fd not suitable for reading - ignore */)
                { // Ignoring some errors common for some device types (including access denied).
                    continue;
                }
                SCXRETHROW(e, L"Unable to handle exception");
            }
            
            SCXCoreLib::SCXHandle<SCXVolumeGroupConfiguration> pBest(0);
            // At address 0x268, an offset is located. Read that.
            // Note: 0x268 is the address of where it reads "LABELONE" + 0x68.
            size_t basic_offset = *(reinterpret_cast<size_t*>(buf+0x268));
            if (basic_offset >= sizeof(buf) || 0 == basic_offset || 0 != (basic_offset%512))
            { // Sanity checks failed.
                continue;
            }
            // At that offset we find another offset that will point us to the
            // data we are after.
            size_t offset = (*(reinterpret_cast<size_t*>(buf+basic_offset+0x28))) + basic_offset;
            if (offset >= sizeof(buf) || 0 == offset || 0 != (offset%512))
            { // Sanity checks failed.
                continue;
            }
            std::istringstream fs(buf+offset);
            try 
            {
                pBest = new SCXVolumeGroupConfiguration(fs);
                if (pBest->GetSeqNo() == 0)
                {
                    pBest = 0;
                }
            }
            catch (SCXVolumeGroupConfigurationException&)
            {
                // Since we are scanning files, parse errors are ok.
                // Seperate from SCXException since we might wanna handle 
                // parse errors differently in the future.
            }
            catch( SCXCoreLib::SCXException&)
            {
                // Since we are scanning we ignore all types of SCXExceptions.
            }

            if (0 != pBest)
            {
                std::wstring vg, lv;
                try
                {
                    pBest->MapPartition(files[i].Get(), vg, lv);
                    if ((vg == volumeGroup) && (lv ==logicalVolume))
                    {
                        result.push_back(files[i].Get());
                    }
                }
                catch (SCXVolumeGroupConfigurationException&)
                {
                    // Since we are scanning files, parse errors are ok.
                    // Seperate from SCXException since we might wanna handle
                    // parse errors differently in the future.
                }
                catch( SCXCoreLib::SCXException&)
                {
                    // Since we are scanning we ignore all types of SCXExceptions.
                }
            }
        }
        return result;
    }
#if defined(PF_DISTRO_REDHAT) && (PF_MAJOR==4)
    /*----------------------------------------------------------------------------*/
    /**
       Returns true if path is likely to be a partition.
       \param[in] path A path to check.
       \returns true if the path is likely to be a partition path. Otherwise false.

       This method uses the assumption that partitions always end with a number of optional 
       letters and a number of digits. For it to be a partion there must be a file existing
       with a prefix after removing numbers and maybe a characters.
    */
    bool SCXVolumeGroupConfiguration::IsPartition(const std::wstring& path)
    {
        SCXCoreLib::SCXFilePath filePath(path);
        std::wstring filename = filePath.GetFilename();
        if (filename.length() == 0)
        {
            return false;
        }
        // Remove digits and test.
        while (filename.length() > 0 && filename.find_last_of(L"0123456789") == (filename.length()-1))
        {
            filename.resize(filename.length()-1);
        }
        if (filename != filePath.GetFilename())
        {
            filePath.SetFilename(filename);
            if (SCXCoreLib::SCXFileInfo(filePath).PathExists())
            {
                return true;
            }
        }
        // remove characters and test.
        while (filename.length() > 0 && filename.find_last_of(L"0123456789") != (filename.length()-1))
        {
            filename.resize(filename.length()-1);
        }
        if (filename.length() > 0)
        {
            filePath.SetFilename(filename);
            if (SCXCoreLib::SCXFileInfo(filePath).PathExists())
            {
                return true;
            }
        }
        return false;
    }
#endif
} /* name space SCXSystemLib */
/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
