/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file

    \brief       Contains the definition of a simple log mediator class.

    \date        2008-07-18 11:44:22

*/
/*----------------------------------------------------------------------------*/
#ifndef SCXLOGMEDIATORSIMPLE_H
#define SCXLOGMEDIATORSIMPLE_H

#include "scxlogmediator.h"
#include "scxlogbackend.h"
#include <scxcorelib/scxhandle.h>
#include <set>

namespace SCXCoreLib
{
    /*----------------------------------------------------------------------------*/
    /**
        Simple implementation of the log mediator interface.
        This class implements a synchronous solution causing the calling thread
        to wait for logging to complete.
    */
    class SCXLogMediatorSimple : public SCXLogMediator
    {
    private:
        /*----------------------------------------------------------------------------*/
        /**
            A strict ordering is needed by the set template.
        */
        struct HandleCompare
        {
            /*----------------------------------------------------------------------------*/
            /**
                Compares two scxhandles.
                It actually compares the addresses of the data that is pointed to by the
                handles. This is of course random, but the exact order is not important in
                this implementation. Only that there _is_ a strict order.
                \param[in] h1 First handle to compare.
                \param[in] h2 Second handle to compare.
                \returns true if h1 is less than h2.
            */
            bool operator()(const SCXHandle<SCXLogItemConsumerIf> h1, const SCXHandle<SCXLogItemConsumerIf> h2) const
            {
                return h1.GetData() < h2.GetData();
            }
        };

        typedef std::set<SCXHandle<SCXLogItemConsumerIf>, HandleCompare> ConsumerSet; //!< Defines a set of consumers.

    public:
        SCXLogMediatorSimple();
        explicit SCXLogMediatorSimple(const SCXThreadLockHandle& lock);

        virtual void LogThisItem(const SCXLogItem& item);
        virtual SCXLogSeverity GetEffectiveSeverity(const std::wstring& module) const;
        virtual bool RegisterConsumer(SCXHandle<SCXLogItemConsumerIf> consumer);
        virtual bool DeRegisterConsumer(SCXHandle<SCXLogItemConsumerIf> consumer);
        virtual void HandleLogRotate();

        const std::wstring DumpString() const;
    private:
        SCXThreadLockHandle m_lock; //!< Thread lock synchronizing access to internal data.
        ConsumerSet m_Consumers; //!< Set of currently subscribed consumers.
    };
}

#endif /* SCXLOGMEDIATORSIMPLE_H */
/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
