/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file

    \brief       Implementation for a log severity filter.

    \date        2008-07-24 15:43:21

*/
/*----------------------------------------------------------------------------*/

#include "scxlogseverityfilter.h"
#include "scxlogitem.h"

namespace SCXCoreLib
{
    /*----------------------------------------------------------------------------*/
    /**
        Default constructor.
    */
    SCXLogSeverityFilter::SCXLogSeverityFilter() :
        m_DefaultSeverity(eNotSet)
    {
    }

    /*----------------------------------------------------------------------------*/
    /**
        Determine if log item passes the severity filter.
        \param[in] item Item to filter.
        \returns true if item passes the filter.
    */
    bool SCXLogSeverityFilter::IsLogable(const SCXLogItem& item) const
    {
        if (eNotSet == item.GetSeverity())
        {
            return false;
        }
        SCXLogSeverity threshold = GetSeverityThreshold(item.GetModule());
        if (eNotSet == threshold)
        {
            return false;
        }
        
        return item.GetSeverity() >= threshold;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Get severity threshold for a module.
        The severity is either explicitly stated for this module, or inherited
        from a parent module. If no severity is set that affects this module,
        then eNotSet is returned.

        \param[in] module Module to get severity threshold for.
        \returns severity The severity threshold for the module.
    */
    SCXLogSeverity SCXLogSeverityFilter::GetSeverityThreshold(const std::wstring& module) const
    {
        /* Example: To get the the effective severity for the module
           scx.core.common.pal.system.common.entityenumeration, first look for
           scx.core.common.pal.system.common.entityenumeration in the map,
           then try scx.core.common.pal.system.common, then try
           scx.core.common.pal.system, then scx.core.common.pal, and so on until we
           reach the root element.
        */
        SeverityMap::const_iterator pos;
        std::wstring effective_module(module);
        std::wstring::size_type dotpos;

        for (;;) {
            pos = m_ModuleMap.find(effective_module);
            if (pos != m_ModuleMap.end()) {
                // Found an effective severity. This is probably it. However,
                // there is an exception to the inheritance rule. If the severity
                // is eHysterical and this is not the original module that 
                // the method was invoked on, then we have to go higher in the
                // inheritance structure.
                if (eHysterical != pos->second || effective_module == module ) { return pos->second; }
            }

            dotpos = effective_module.rfind(L'.');
            if (dotpos == std::wstring::npos) {
                // We've just tested the top level, "scx" in the example.
                break;
            }
            
            // Erase part from dot to end of string
            effective_module.erase(dotpos);
        }
        // We've tried our module name and all superior modules without result
        // Now return the default
        return m_DefaultSeverity;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Set severity threshold for a module.

        \param[in] module Module to set severity threshold for.
        \param[in] severity The severity threshold to set.
        \returns true if the severity filter was actually changed as a result of
        this method call.
    */
    bool SCXLogSeverityFilter::SetSeverityThreshold(const std::wstring& module, SCXLogSeverity severity)
    {
        if (module.empty())
        {
            if (m_DefaultSeverity != severity && eHysterical != severity)
            {
                m_DefaultSeverity = severity;
                return true;
            }
        }
        else
        {
            SeverityMap::iterator iter = m_ModuleMap.lower_bound(module);
            if ((m_ModuleMap.end() != iter) && !(m_ModuleMap.key_comp()(module, iter->first)))
            {
                if (iter->second != severity)
                {
                    iter->second = severity;
                    return true;
                }
            }
            else
            {
                m_ModuleMap.insert(iter, std::pair<const std::wstring, SCXLogSeverity>(module, severity));
                return true;
            }
        }
        return false;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Unset severity threshold for a module.

        \param[in] module Module to clear severity threshold for.
        \returns true if the severity filter was actually changed as a result of
        this method call.
    */
    bool SCXLogSeverityFilter::ClearSeverityThreshold(const std::wstring& module)
    {
        if (module.empty())
        {
            if (m_DefaultSeverity != eNotSet)
            {
                m_DefaultSeverity = eNotSet;
                return true;
            }
        }
        else
        {
            return (1 == m_ModuleMap.erase(module));
        }
        return false;
    }

    /**
        Get the minimum log severity threshold used for any module in this filter.
        \returns Minimum severity threshold used for this filter.
    */
    SCXLogSeverity SCXLogSeverityFilter::GetMinActiveSeverityThreshold() const
    {
        SCXLogSeverity s = m_DefaultSeverity;

        for (SeverityMap::const_iterator iter = m_ModuleMap.begin();
             iter != m_ModuleMap.end();
             ++iter)
        {
            if (iter->second < s)
            {
                s = iter->second;
            }
        }
        return s;
    }
} /* namespace SCXCoreLib */
/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
