/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file        

    \brief       Utilities to implement DumpString methods 
    
    \date        2007-12-04 14:20:00

    Every class should have a DumpString method for debugging purposes. This file offers
    functionality to ease the implementation of those, as well as making the return values coherent. 
    
*/
/*----------------------------------------------------------------------------*/
#ifndef SCXDUMPSTRING_H
#define SCXDUMPSTRING_H
 
#include <string>
#include <sstream>
#include <vector>
/*----------------------------------------------------------------------------*/
/**
    Constructs a return value for DumpString        
*/
class SCXDumpStringBuilder {
public:
    SCXDumpStringBuilder(std::string classname);

    SCXDumpStringBuilder &Text(const std::string &name, const std::wstring &value);

    /*----------------------------------------------------------------------------*/
    //! Appends a scalar value
    //! \param[in]  name       Name (role) of the value
    //! \param[in]  value      Value to be appended
    //! \returns    *this
    template <class T>
    inline SCXDumpStringBuilder &Scalar(const std::string &name, const T value) {
        m_stream << L" " << name.c_str() << L"=" << value;
        return *this;
    }
    
    /*----------------------------------------------------------------------------*/
    //! Appends the DumpString value of an instance
    //! \param[in]  name       Name (role) of the instance
    //! \param[in]  instance   Instance whose DumpString is to be appended
    //! \returns    *this
    template <class T>
    inline SCXDumpStringBuilder &Instance(const std::string &name, const T &instance) {
        m_stream << L" " << name.c_str() << L"=[" << instance.DumpString() << L"]";
        return *this;
    }

    /*----------------------------------------------------------------------------*/
    //! Appends values
    //! \param[in]  name        Name (role) of the values
    //! \param[in]  values      Values to be appended
    //! \returns    *this
    template <class T>
    SCXDumpStringBuilder &Scalars(const std::string &name, const std::vector<T> &values) {
        m_stream << L" " << name.c_str() << L"={";
        typename std::vector<T>::const_iterator i = values.begin();
        if (i != values.end()) {
            m_stream << *i;
            ++i;
            while (i != values.end()) {
                m_stream << L" " << *i;            
                ++i;
            }
        }
        m_stream << L"}";
        return *this;
        
    }

    /*----------------------------------------------------------------------------*/
    //! Appends the DumpString value of instances
    //! \param[in]  name        Name (role) of the value
    //! \param[in]  instances   Instances whose DumpStrings are to be appended
    //! \returns    *this
    template <class T>
    SCXDumpStringBuilder &Instances(const std::string &name, const std::vector<T> &instances) {
        m_stream << L" " << name.c_str() << L"={";
        typename std::vector<T>::const_iterator i = instances.begin();
        if (i != instances.end()) {
            m_stream << L"[" << i->DumpString() << L"]";
            ++i;
            while (i != instances.end()) {
                m_stream << L" [" << i->DumpString() << L"]";            
                ++i;
            }
        }
        m_stream << L"}";
        return *this;
        
    }

    std::wstring Str() const;
    operator std::wstring () const;

private:
    //! String stream to hold string
    std::wostringstream m_stream;
};


#endif /* SCXDUMPSTRING_H */
/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
