/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file        

    \brief       Declares a PAL to handle RAID configurations.
    
    \date        2008-03-28 09:50:00

*/
/*----------------------------------------------------------------------------*/
#ifndef SCXRAID_H
#define SCXRAID_H

#include <scxcorelib/scxhandle.h>
#include <scxcorelib/scxfilepath.h>
#include <scxcorelib/scxexception.h>

#include <string>
#include <vector>
#include <map>

namespace SCXSystemLib
{
/*----------------------------------------------------------------------------*/
/**
    An interface for Raid configuration parsers.
*/
    class SCXRaidCfgParser
    {
    public:
        /*----------------------------------------------------------------------------*/
        /**
           Enumeration to indicate what kind of configuration line was parsed.
        */
        enum ParsedLineType
        {
            eSCX_RAID_CFG_LINE_TYPE_UNKNOWN = 0,  //!< unknown device type 
            eSCX_RAID_CFG_LINE_TYPE_STRIPE,       //!< striped (RAID0) and/or concat
            eSCX_RAID_CFG_LINE_TYPE_MIRROR,       //!< mirror (RAID1)
            eSCX_RAID_CFG_LINE_TYPE_TRANS,        //!< trans (master and logging)
            eSCX_RAID_CFG_LINE_TYPE_HOT_SPARE,    //!< hot spare pool
            eSCX_RAID_CFG_LINE_TYPE_RAID,         //!< raid (RAID5)
            eSCX_RAID_CFG_LINE_TYPE_MAX           //!< enum max marker
        };

        /** Default virtual constructor */
        virtual ~SCXRaidCfgParser() { }
        
        /*----------------------------------------------------------------------------*/
        /**
           Get the raid configuration path.

           \returns The path to raid configuration.
        */
        virtual const SCXCoreLib::SCXFilePath& GetConfPath() const = 0;

        /*----------------------------------------------------------------------------*/
        /**
           Prepare configuration lines for parsing.

           \param lines A vector of lines to prepare. Lines will be changed and removed as needed.

           Typicaly this method removes comments, empty lines and merges lines when needed.
        */
        virtual void PrepareLines(std::vector<std::wstring>& lines) = 0;

        /*----------------------------------------------------------------------------*/
        /**
           Parse a configuration line.

           \param[in] line Line to parse.
           \param[out] md Meta device parsed.
           \param[out] devices A list of devices part of the meta device.
           \param[out] options A map of options parsed.
           \returns The type of configuration line parsed.
        */
        virtual ParsedLineType ParseLine(const std::wstring& line, std::wstring& md, std::vector<std::wstring>& devices, std::map<std::wstring, std::wstring>& options) = 0;
    };

/*----------------------------------------------------------------------------*/
/**
    Default raid configuration parser

    \note Only tested for solaris.
*/
    class SCXRaidCfgParserDefault : public SCXRaidCfgParser
    {
    public:
        virtual const SCXCoreLib::SCXFilePath& GetConfPath() const;
        virtual void PrepareLines(std::vector<std::wstring>& lines);
        virtual ParsedLineType ParseLine(const std::wstring& line, std::wstring& md, std::vector<std::wstring>& devices, std::map<std::wstring, std::wstring>& options);

    protected:
        bool IsHotSpare(const std::wstring& device) const;
        void ParseOptions(std::vector<std::wstring>& words, std::map<std::wstring, std::wstring>& options);
        std::vector<std::wstring> m_hotSpares; //!< A list of seen hot spare names.
    };

/*----------------------------------------------------------------------------*/
/**
    RAID representation.
*/
    class SCXRaid
    {
    public:
        SCXRaid(const SCXCoreLib::SCXHandle<SCXRaidCfgParser>& parser);
        
        void GetMetaDevices(std::vector<std::wstring>& devices) const;
        void GetDevices(const std::wstring& md, std::vector<std::wstring>& devices) const;

        std::wstring DumpString() const;
    private:
        /** Default constructor should not be used - made private */
        SCXRaid() : m_parser(0) { }

        SCXCoreLib::SCXHandle<SCXRaidCfgParser> m_parser; //!< The config parser helper.
        //! A map with meta device to devices mappings.
        std::map<std::wstring, std::vector<std::wstring> > m_devices; 
        //! A map with mirror mappings
        std::map<std::wstring, std::vector<std::wstring> > m_mirrors;
        //! A map of trans devices
        std::map<std::wstring, std::vector<std::wstring> > m_trans;
        //! A map with hot spare mappings (what devices is part of a hot spare)
        std::map<std::wstring, std::vector<std::wstring> > m_hotSpares;
        //! A map with device to hot spare mapping (i.e. what hotspare is part of meta device)
        std::map<std::wstring, std::wstring> m_deviceHotSpare;
    };

    /*----------------------------------------------------------------------------*/
    /**
       A generic exception for raid configuration parsing exceptions.
    */
    class SCXRaidConfigParseErrorException: public SCXCoreLib::SCXException
    {
    public:
        /**
           Ctor
           \param[in] reason Description of error
           \param[in] l      Source code location object

        */
        SCXRaidConfigParseErrorException(const std::wstring& reason, const SCXCoreLib::SCXCodeLocation& l)
            : SCXCoreLib::SCXException(l), m_Reason(reason)
        { }

        std::wstring What() const {
            return L"RAID Config Parsing Error: " + m_Reason;
        }

    protected:
        //! Description of error
        std::wstring   m_Reason;

    };


} /* namespace SCXSystemLib */
#endif /* SCXRAID_H */
/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
