/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
   \file        

   \brief       Defines the volume group configuration parser.
    
   \date        2008-01-17 15:00:00

   This parser assumes a volume group configuration on the following format (Linux LVM2)

   \verbatim
   VolGroupXX {
     <attributes>
     physical_volumes {
       pvN {
         <attributes>
       }
       ...
     }
     logical_volumes {
       LogVolYY {
         segmentZ {
           <attributes>
         }
         <attributes>
       }
       ...
     }
   }
   <additional attributes>
   \endverbatim

   The parser assumes the above format regarding whitespace location (but not ammount)
   and line breaks. I.e. sections (and partial attributes) have there end token on a 
   seperate line.
*/
/*----------------------------------------------------------------------------*/
#ifndef SCXVGCFG_H
#define SCXVGCFG_H

#include <iostream>
#include <vector>
#include <map>

#include <scxcorelib/scxexception.h>
#include <scxcorelib/scxhandle.h>

namespace SCXSystemLib {

    class SCXVolumeGroupConfigurationItem;

    /** Vector of SCXVolumeGroupConfigurationItem. */
    typedef std::vector<SCXCoreLib::SCXHandle<SCXSystemLib::SCXVolumeGroupConfigurationItem> > SCXVolumeGroupConfigurationItemVector;
    /** A string -> string map holding attributes names and values. */
    typedef std::map<std::wstring,std::wstring> SCXVolumeGroupConfigurationItemAttributes;

/*----------------------------------------------------------------------------*/
/**
   Represents a single configuration item (or section).
    
   Items are organized in a hiarchical manner.
*/
    class SCXVolumeGroupConfigurationItem
    {
        friend class SCXVolumeGroupConfiguration;
    public:
        SCXVolumeGroupConfigurationItem();
        virtual ~SCXVolumeGroupConfigurationItem();

        const std::wstring& GetName() const;
        const SCXVolumeGroupConfigurationItemVector& GetSections() const;
        const SCXVolumeGroupConfigurationItemAttributes& GetAttributes() const;
    protected:
        void Parse(std::vector<std::wstring>& lines);
        bool ParseAttribute(std::wstring& line, bool partial);
    private:
        std::wstring m_name; //!< Item name
        SCXVolumeGroupConfigurationItemVector m_sections; //!< List of child sections.
        SCXVolumeGroupConfigurationItemAttributes m_attributes; //!< Item attributes.
        std::wstring m_partialAttributeName; //!< State member holding partial attribute's name. used during parsing.
    };

/*----------------------------------------------------------------------------*/
/**
   Represents a volume group configuration.
    
*/
    class SCXVolumeGroupConfiguration
    {
    public:
        SCXVolumeGroupConfiguration(std::istream& stream);
        virtual ~SCXVolumeGroupConfiguration();

        /*----------------------------------------------------------------------------*/
/**
   Get root of configuration.

   \returns The root item of the configuration.
*/
        const SCXVolumeGroupConfigurationItem& GetRoot() const { return m_root; }

        size_t GetSeqNo() const;
        void MapPartition(const std::wstring& partition, std::wstring& volumeGroup, std::wstring& logicalVolume) const;

        static std::vector<std::wstring> FindPartitions(const std::wstring& volumeGroup, const std::wstring& logicalVolume);
#if defined(PF_DISTRO_REDHAT) && (PF_MAJOR==4)
        static bool IsPartition(const std::wstring& path);
#endif
    protected:
        SCXVolumeGroupConfigurationItem m_root; //!< The configuration item representing the root of the configuration.
    };

/*----------------------------------------------------------------------------*/
/**
   A generic exception from the volume group configuration parser.
    
*/
    class SCXVolumeGroupConfigurationException: public SCXCoreLib::SCXException
    {
    public:
/*----------------------------------------------------------------------------*/
/**
   Constructor

   \param       hint - Reason for throwing exception.
   \param       location - Source code location object

*/
        SCXVolumeGroupConfigurationException(const std::wstring& hint, const SCXCoreLib::SCXCodeLocation& location)
            : SCXCoreLib::SCXException(location), m_hint(hint)
        { }
/*----------------------------------------------------------------------------*/
/**
    Format reason why configuration is invalid.

    \returns     String describing why volume group configuration was not parsed.

*/
        std::wstring What() const { return L"Unable to parse volume group configuration: " + m_hint; }
    protected:
        std::wstring m_hint; //!< Hint to reason for exception.
    };

} /* namespace SCXSystemLib */
#endif /* SCXVGCFG_H */
/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
