/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file
 
    \brief     Main implementation file for Log File Provider
 
    \date      2008-0-08 09:35:36
 

*/
/*----------------------------------------------------------------------------*/

#include <scxcorelib/scxcmn.h>

#include <fstream>
#include <errno.h>

#include <scxcorelib/scxexception.h>
#include <scxcorelib/scxlog.h>
#include <scxcorelib/scxmath.h>
#include <scxcorelib/scxuser.h>
#include <scxcorelib/stringaid.h>
#include <scxcorelib/scxstream.h>
#include <scxcorelib/scxregex.h>
#include <scxcorelib/scxfile.h>

#include <scxproviderlib/scxprovidercapabilities.h>

#include "logfileprovider.h"
#include "../meta_provider/startuplog.h"

using namespace SCXProviderLib;
using namespace SCXCoreLib;
using namespace std;

namespace SCXCore {

    const SCXCoreLib::SCXPatternFinder::SCXPatternCookie LogFileProvider::s_patternID = 1;
    const std::wstring LogFileProvider::s_pattern = L"SELECT * FROM SCX_LogFileRecord WHERE FileName=%PATH";
    const std::wstring LogFileProvider::s_patternParameter = L"PATH";
    const unsigned int cMaxRows = 5000;   //!< max number of log rows return limit 
    const unsigned int cMaxMatchedRows = 500;   //!< max number of matched log rows return limit, 1000 rows from scx log file does not work, 750 does

    /*----------------------------------------------------------------------------*/
    /* LogFileProvider::LogFilePositionRecord                                     */
    /*----------------------------------------------------------------------------*/

    /*----------------------------------------------------------------------------*/
    /**
        Constructor
        Creates a new LogFilePositionRecord with default values.

        \param[in] logfile Log file for this record.
        \param[in] qid Q ID of this record.
        \param[in] persistMedia Used to inject persistence media to use for persisting this record. 
    */
    LogFileProvider::LogFilePositionRecord::LogFilePositionRecord(
        const SCXCoreLib::SCXFilePath& logfile,
        const std::wstring& qid,
        SCXCoreLib::SCXHandle<SCXCoreLib::SCXPersistMedia> persistMedia /* =  SCXCoreLib::GetPersistMedia()*/)
        : m_PersistMedia(persistMedia),
          m_LogFile(logfile),
          m_Pos(0),
          m_StIno(0),
          m_StSize(0)
    {
        SCXUser user;
        m_IdString = L"LogFileProvider_" + user.GetName() + logfile.Get() + qid;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Get the log file associated with this record.
        \returns File path to log file.
    */
    const SCXFilePath& LogFileProvider::LogFilePositionRecord::GetLogFile() const
    {
        return m_LogFile;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Get current file position for the logfile and qid.
        \returns Current file position.
    */
    std::streamoff LogFileProvider::LogFilePositionRecord::GetPos() const
    {
        return m_Pos;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Set current file position for the logfile and qid.
        \param[in] pos Current file position.
    */
    void LogFileProvider::LogFilePositionRecord::SetPos(std::streamoff pos)
    {
        m_Pos = pos;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Get the st_ino filed of a stat struct for this file.
        \returns st_ino field of a stat structure.
    */
    scxulong LogFileProvider::LogFilePositionRecord::GetStatStIno() const
    {
        return m_StIno;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Set the st_ino filed of a stat struct for this file.
        \param[in] st_ino st_ino field of a stat structure.
    */
    void LogFileProvider::LogFilePositionRecord::SetStatStIno(scxulong st_ino)
    {
        m_StIno = st_ino;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Get the st_size filed of a stat struct for this file.
        \returns st_size field of a stat structure.
    */
    scxulong LogFileProvider::LogFilePositionRecord::GetStatStSize() const
    {
        return m_StSize;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Set the st_size filed of a stat struct for this file.
        \param[in] st_size st_size field of a stat structure.
    */
    void LogFileProvider::LogFilePositionRecord::SetStatStSize(scxulong st_size)
    {
        m_StSize = st_size;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Persist data
    */
    void LogFileProvider::LogFilePositionRecord::Persist()
    {
        if (static_cast<scxulong>(m_Pos) > m_StSize)
        {
            m_StSize = static_cast<scxulong>(m_Pos);
        }
        SCXHandle<SCXPersistDataWriter> pwriter = m_PersistMedia->CreateWriter(m_IdString, 0);
        pwriter->WriteValue(L"Pos", SCXCoreLib::StrFrom(m_Pos));
        pwriter->WriteStartGroup(L"Stat");
        pwriter->WriteValue(L"StIno", SCXCoreLib::StrFrom(m_StIno));
        pwriter->WriteValue(L"StSize", SCXCoreLib::StrFrom(m_StSize));
        pwriter->WriteEndGroup();
        pwriter->DoneWriting();
    }

    /*----------------------------------------------------------------------------*/
    /**
        Recover persisted data
        \returns false if no data had previously been persisted.
    */
    bool LogFileProvider::LogFilePositionRecord::Recover()
    {
        try
        {
            SCXHandle<SCXPersistDataReader> preader = m_PersistMedia->CreateReader(m_IdString);
            if (0 != preader->GetVersion())
            {
                // Wrong version. Just ignore. It will be re-persisted later.
                return false;
            }

            m_Pos = SCXCoreLib::StrToULong(preader->ConsumeValue(L"Pos"));
            preader->ConsumeStartGroup(L"Stat");
            m_StIno = SCXCoreLib::StrToULong(preader->ConsumeValue(L"StIno"));
            m_StSize = SCXCoreLib::StrToULong(preader->ConsumeValue(L"StSize"));
            preader->ConsumeEndGroup();
            return true;
        }
        catch (SCXNotSupportedException&)
        {
            // Data is corrupt (a value could not be parsed). Just ignore. It will be re-persisted later.
            return false;
        }
        catch (PersistUnexpectedDataException&)
        {
            // Data is corrupt. Just ignore. It will be re-persisted later.
            return false;
        }
        catch (PersistDataNotFoundException&)
        {
            // No persisted data found.
            return false;
        }
    }

    /*----------------------------------------------------------------------------*/
    /**
        Remove persisted data
        \returns false if no data had previously been persisted.
    */
    bool LogFileProvider::LogFilePositionRecord::UnPersist()
    {
        try
        {
            m_PersistMedia->UnPersist(m_IdString);
        }
        catch (PersistDataNotFoundException&)
        {
            // No persisted data found.
            return false;
        }
        return true;
    }

    /*----------------------------------------------------------------------------*/
    /* LogFileProvider::LogFileStreamPositioner                                   */
    /*----------------------------------------------------------------------------*/

    /*----------------------------------------------------------------------------*/
    /**
        Constructor

        \param[in] logfile Log file for this record.
        \param[in] qid Q ID of this record.
        \param[in] persistMedia Used to inject persistence media to use for persisting this record. 
        \throws SCXFilePathNotFoundException if log file does not exist.
    */
    LogFileProvider::LogFileStreamPositioner::LogFileStreamPositioner(
        const SCXCoreLib::SCXFilePath& logfile,
        const std::wstring& qid,
        SCXCoreLib::SCXHandle<SCXCoreLib::SCXPersistMedia> persistMedia /* =  SCXCoreLib::GetPersistMedia()*/)
        : m_Record(0),
          m_Stream(0),
          m_log(SCXLogHandleFactory::GetLogHandle(L"scx.core.providers.logfileprovider.logfilestreampositioner"))
    {
        m_Record = new LogFilePositionRecord(logfile, qid, persistMedia);
        m_Stream = SCXFile::OpenWFstream(logfile, std::ios_base::in);
        if ( ! m_Record->Recover())
        {
            // First time (i.e when no persisted data exists) we just go to end of file.
            SCX_LOGTRACE(m_log, L"OpenStream " + m_Record->GetLogFile().Get()
                         + L" - First time - Seek to end");
            m_Stream->seekg(0, std::ios_base::end);
            std::streamoff pos = m_Stream->tellg();
            m_Record->SetPos(pos);
            SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider OpenStream last pos = ", pos));
        }
        else
        {
            if ( ! IsFileNew())
            {
                // File has not wrapped so we seek to last position.
                SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider OpenLogFile " + m_Record->GetLogFile().Get()
                                              + L"- Seek to: ", m_Record->GetPos()));
                m_Stream->seekg(0, std::ios_base::end);
                std::streamoff pos = m_Stream->tellg();
                SCXFile::SeekG(*m_Stream, m_Record->GetPos());
                SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider OpenStream save last pos = ", pos));
                m_Record->SetPos(pos);
            }
        }
        UpdateStatData();
    }

    /*----------------------------------------------------------------------------*/
    /**
        Return a stream pointing at the correct reading position.

        \returns       Handle to stream opened at the correct position.
    */
    SCXHandle<std::wfstream> LogFileProvider::LogFileStreamPositioner::GetStream()
    {
        return m_Stream;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Save the state of a logfile.
    */
    void LogFileProvider::LogFileStreamPositioner::PersistState()
    {
        std::streamoff pos = m_Stream->tellg();
        SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider PersistState() - pos = ", pos));
        // Never persist -1, happens on some platforms (AIX) when reading past end of file
        // Real end of file position has been saved when opening the file
        if (pos > 0)
        {
            m_Record->SetPos(pos);
        }
        m_Record->Persist();
    }

    /*----------------------------------------------------------------------------*/
    /**
        Check if the log file is actually a new file with the same name.
        This will happen if the file has wrapped since last update for example.

        \returns true if the file is actually a new file.
        \throws SCXFilePathNotFoundException if log file does not exist.
    */
    bool LogFileProvider::LogFileStreamPositioner::IsFileNew() const
    {
        SCXFileSystem::SCXStatStruct statstruct;
        SCXFileSystem::Stat(m_Record->GetLogFile(), &statstruct);
         
        // If inode has changed it is a new file
        if (statstruct.st_ino != m_Record->GetStatStIno())
        {
            SCX_LOGTRACE(m_log, L"IsNewFile - inode changed - new file");
            return true;
        }
        
        // If the new size is smaller it is a new file
        if ((scxulong) statstruct.st_size < m_Record->GetStatStSize())
        {
            SCX_LOGTRACE(m_log, L"IsNewFile - size smaller - new file");
            return true;
        }

        SCX_LOGTRACE(m_log, L"IsNewFile - not new file");
        
        return false;
    }

    /*----------------------------------------------------------------------------*/
    /**
        Update stat fields in record with values read from disk.
        \throws SCXFilePathNotFoundException if log file does not exist.
    */
    void LogFileProvider::LogFileStreamPositioner::UpdateStatData()
    {
        SCXFileSystem::SCXStatStruct statstruct;
        SCXFileSystem::Stat(m_Record->GetLogFile(), &statstruct);
         
        m_Record->SetStatStIno(statstruct.st_ino);
        m_Record->SetStatStSize(statstruct.st_size);
    }

    /*----------------------------------------------------------------------------*/
    /**
        Provide CMPI interface for this class
       
        The class implementation (concrete class) is LogFileProvider and the name of the 
        provider in CIM registration terms is SCX_LogFileProvider.
    */
    SCXProviderDef(LogFileProvider, SCX_LogFileProvider)

    /*----------------------------------------------------------------------------*/
    /**
       Default constructor
       
       The Singleton thread lock will be held during this call.
       
    */
    LogFileProvider::LogFileProvider() : BaseProvider(L"scx.core.providers.logfileprovider")
    {
        LogStartup();
        SCX_LOGTRACE(m_log, L"LogFileProvider constructor");

        m_persistMedia = GetPersistMedia();
        m_cqlPatterns.RegisterPattern(s_patternID, s_pattern);
    }
    
    /*----------------------------------------------------------------------------*/
    /**
       Destructor
    */
    LogFileProvider::~LogFileProvider()
    {
        // Do not log here since when this destructor is called the objects neccesary for logging might no longer be alive
    }

    /*----------------------------------------------------------------------------*/
    /**
       Registration of supported capabilities

       Callback from the BaseProvider in which the provider registers the supported
       classes and methods. The registrations should match the contents of the 
       MOF files exactly.

    */
    void LogFileProvider::DoInit()
    {
        SCX_LOGTRACE(m_log, L"LogFileProvider::DoInit()");

        m_ProviderCapabilities.RegisterCimClass(eSCX_LogFile, L"SCX_LogFile");
        m_ProviderCapabilities.RegisterCimClass(eSCX_LogFileRecord, L"SCX_LogFileRecord");
        m_ProviderCapabilities.RegisterCimMethod(eSCX_LogFile, eGetMatchedRowsMethod, L"GetMatchedRows");
    }
    

    /*----------------------------------------------------------------------------*/
    /**
        Provide a way for pal layer to do cleanup. Stop all threads etc.
    */
    void LogFileProvider::DoCleanup() 
    {
        SCX_LOGTRACE(m_log, L"LogFileProvider::DoCleanup");

        m_ProviderCapabilities.Clear();
    }

    /*----------------------------------------------------------------------------*/
    /**
       Enumerate instance names

       \param[in]   callContext Context details of this request
       \param[out]  instances   Collection of instances with key properties

    */
    void LogFileProvider::DoEnumInstanceNames(const SCXCallContext& callContext, 
                                              SCXInstanceCollection& instances)
    { 
        SCX_LOGTRACE(m_log, L"LogFileProvider DoEnumInstanceNames");

        SupportedCimClasses cimtype = static_cast<SupportedCimClasses>(m_ProviderCapabilities.GetCimClassId(callContext.GetObjectPath()));
        
        if (cimtype != eSCX_LogFile)
        {
            throw SCXNotSupportedException(L"LogFileProvider Enumerate not for LogFile class", SCXSRCLOCATION);
        }

        for(size_t i=0; i<m_files.size(); i++)
        {
            SCXInstance inst;

            SCXProperty name_prop(L"Name", m_files[i].name);
            inst.AddKey(name_prop);

            instances.AddInstance(inst);
        }
    }
    
    /*----------------------------------------------------------------------------*/
    /**
       Enumerate instances

       \param[in]     callContext                 Details of the client request
       \param[out]    instances                   Collection of instances

       \throws        SCXInternalErrorException   If instances in list are not CPUInstance

    */
    void LogFileProvider::DoEnumInstances(const SCXCallContext& callContext, 
                                          SCXInstanceCollection& instances)  
    { 
        SCX_LOGTRACE(m_log, L"LogFileProvider DoEnumInstances");

        DoEnumInstanceNames(callContext, instances);
    }

    /*----------------------------------------------------------------------------*/
    /**
       Get filename from query

       \param[in]     query        The CQL/WQL query to parse

       \returns       Log file name

       \throws        SCXNotSupportedException   If query not on recognized format
       \throws        SCXInternalErrorException  If an unexpected pattern was matched.

    */
    std::wstring LogFileProvider::GetFileName(const std::wstring& query)
    {
        SCXCoreLib::SCXPatternFinder::SCXPatternCookie id = 0;
        SCXCoreLib::SCXPatternFinder::SCXPatternMatch param;
        if ( ! m_cqlPatterns.Match(query, id, param))
        {
            throw SCXNotSupportedException(L"LogFileProvider Query not on format: " + s_pattern, SCXSRCLOCATION);
        }
        if (id != s_patternID || param.end() == param.find(s_patternParameter))
        {
            throw SCXInternalErrorException(L"Wrong pattern matched!", SCXSRCLOCATION);
        }
        SCX_LOGTRACE(m_log, L"LogFileProvider GetFileName: " + param.find(s_patternParameter)->second);

        return param.find(s_patternParameter)->second;
    }

    /*----------------------------------------------------------------------------*/
    /**
       Get SCXLogFile class for a given log file name

       \param[in]     filename        Name of log file

       \returns       Pointer to corresponding SCXLogFile class or NULL if none found

    */
    LogFileProvider::SCXLogFile* LogFileProvider::GetLogFile(const std::wstring& filename)
    {
        for(size_t i=0; i<m_files.size(); i++)
        {
            if (m_files[i].name == filename)
            {
                SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider GetLogFile found filename - ", filename));
                return &m_files[i];
            }
        }

        SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider GetLogFile did NOT find filename - ", filename));
        return NULL;
    }


    /*----------------------------------------------------------------------------*/
    /**
       Check if the log file has wrapped and we need to start reading from line 1 again

       \param[in]     oldstatinfo     Old saved stat info
       \param[in]     newstatinfo     New stat info

       \returns       true if file has wrapped, otherwise false

    */
    bool LogFileProvider::CheckFileWrap(const struct stat64& oldstatinfo, const struct stat64& newstatinfo)
    {
        // If inode has changed it is a new file
        if (oldstatinfo.st_ino != newstatinfo.st_ino)
        {
            SCX_LOGTRACE(m_log, L"LogFileProvider CheckFileWrap - inode changed - new file");
            return true;
        }
        
        // If the new size is smaller it is a new file
        if (oldstatinfo.st_size > newstatinfo.st_size)
        {
            SCX_LOGTRACE(m_log, L"LogFileProvider CheckFileWrap - size smaller - new file");
            return true;
        }

        SCX_LOGTRACE(m_log, L"LogFileProvider CheckFileWrap - inode not changed and size not smaller - not new file");
        return false;
    }

    /*----------------------------------------------------------------------------*/
    /**
       Execute Query

       \param[in]     callContext                 Details of the client request
       \param[out]    instances                   Collection of instances
       \param[in]     query                       The WQL/CQL query to execute
       \param[in]     language                    Type of query (CQL/WQL)

       \throws        SCXNotSupportedException    If unhandled class

    */
    void LogFileProvider::DoExecQuery(const SCXProviderLib::SCXCallContext& callContext, SCXInstanceCollection& instances,
                                      std::wstring query, std::wstring language)
    {
        SCX_LOGTRACE(m_log, wstring(L"LogFileProvider DoExecQuery - ").append(query).append(L" - ").append(language));

        SupportedCimClasses cimtype = static_cast<SupportedCimClasses>(m_ProviderCapabilities.GetCimClassId(callContext.GetObjectPath()));
        
        if (cimtype != eSCX_LogFileRecord)
        {
            throw SCXNotSupportedException(L"LogFileProvider Query not for LogFileRecord class", SCXSRCLOCATION);
        }

        wstring filename = GetFileName(query);

        SCXLogFile* logfilep = GetLogFile(filename);

        if (logfilep == NULL)
        {
            SCXLogFile logfile;
            logfile.name = filename;

            // Save end of file position
            wifstream file(StrToMultibyte(filename).c_str());
            if (SCXStream::IsGood(file))
            {
                file.seekg(0, ios_base::end);
                logfile.pos = file.tellg();
                logfile.statvalid = (stat64(StrToMultibyte(filename).c_str(), &logfile.statinfo) == 0);
                if (!logfile.statvalid)
                {
                    SCX_LOGWARNING(m_log, StrAppend(L"LogFileProvider DoExecQuery - first stat64 failed - ", errno));
                }
            }
            else
            {
                logfile.pos = 0;
                logfile.statvalid = false;
            }
            SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider DoExecQuery - Set first pos to: ", logfile.pos));
            
            m_files.push_back(logfile);
        }
        else
        {
            wfstream logfile(StrToMultibyte(logfilep->name).c_str(), std::ios::in);
            unsigned int rows = 0;

            if (SCXStream::IsGood(logfile))
            {
                streamoff pos = logfilep->pos;
                struct stat64 statinfo = logfilep->statinfo;
                bool statvalid = logfilep->statvalid;

                logfilep->statvalid = (stat64(StrToMultibyte(logfilep->name).c_str(), &logfilep->statinfo) == 0);

                if (!logfilep->statvalid)
                {
                    SCX_LOGWARNING(m_log, StrAppend(L"LogFileProvider DoExecQuery - stat64 failed - ", errno));
                }

                // Save new end of file position
                logfile.seekg(0, ios_base::end);
                logfilep->pos = logfile.tellg();
                SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider DoExecQuery - Set pos to: ", logfilep->pos));

                if (pos > 0 && statvalid && logfilep->statvalid && CheckFileWrap(statinfo, logfilep->statinfo))
                {
                    pos = 0;
                }

                // Start reading from old end of file position
                SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider DoExecQuery - Seek to: ", pos));
                SCXFile::SeekG(logfile, pos);

                // Get all rows added to file since last call
                while (rows < cMaxRows && SCXStream::IsGood(logfile))
                {
                    wstring line;
                    SCXStream::NLF nlf;
                    SCXInstance inst;
                    pos = logfile.tellg();

                    rows++;

                    SCX_LOGHYSTERICAL(m_log, StrAppend(L"LogFileProvider DoExecQuery - Reading row: ", rows));

                    SCXStream::ReadLine(logfile, line, nlf);

                    SCXProperty row_prop(L"RecordData", line);
                    inst.AddProperty(row_prop);
                    SCXProperty filename_prop(L"FileName", filename);
                    inst.AddKey(filename_prop);
                    SCXProperty filepos_prop(L"FilePosition", (scxulong)pos);
                    inst.AddKey(filepos_prop);

                    if (rows >= cMaxRows)
                    {
                        if (SCXStream::IsGood(logfile))
                        {
                            logfilep->pos = logfile.tellg();
                            SCX_LOGINFO(m_log, StrAppend(L"LogFileProvider DoExecQuery - Breaking after max number of rows : ", cMaxRows));
                            SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider DoExecQuery - Reset pos to: ", logfilep->pos));
                        }
                    }

                    instances.AddInstance(inst);

                }
            }
            else
            {
                logfilep->pos = 0;
                logfilep->statvalid = false;
                SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider DoExecQuery - File not found: ", filename));
            }

            SCX_LOGTRACE(m_log, StrAppend(L"LogFileProvider DoExecQuery - Returning lines: ", rows));

        }
    }

    /*----------------------------------------------------------------------------*/
    /**
       Helper structure to hold a regular expression together with its index
    */
    struct SCXRegexWithIndex
    {
        size_t index;     //!< index for regular expression
        SCXHandle<SCXRegex> regex;  //!< the regular expression
    };

    /*----------------------------------------------------------------------------*/
    /**
        Invoke a method on an instance

        \param[in]     callContext Keys indicating instance to execute method on
        \param[in]     methodname  Name of method called
        \param[in]     args        Arguments provided for method call
        \param[out]    outargs     Output arguments
        \param[out]    result      Result value
    */
    void LogFileProvider::DoInvokeMethod(
        const SCXCallContext& callContext,
        const std::wstring& methodname,
        const SCXArgs& args,
        SCXArgs& outargs,
        SCXProperty& result)
    {
        SCX_LOGTRACE(m_log, L"SCXLogFileProvider DoInvokeMethod");

        // Check that call is made on correct class
        SupportedCimClasses cimtype = static_cast<SupportedCimClasses>(m_ProviderCapabilities.GetCimClassId(callContext.GetObjectPath()));
        
        if (cimtype != eSCX_LogFile)
        {
            throw SCXNotSupportedException(L"LogFileProvider method invoke not for LogFile class", SCXSRCLOCATION);
        }
        
        // Check that call is made to correct method
        SupportedCimMethods cimmethod = static_cast<SupportedCimMethods>(m_ProviderCapabilities.GetCimMethodId(callContext.GetObjectPath(), methodname));
        if (cimmethod == eGetMatchedRowsMethod)
        {
            // Get arguments, and check for correct type
            const SCXProperty* filename_p = args.GetProperty(L"filename");
            const SCXProperty* regexps_p = args.GetProperty(L"regexps");
            const SCXProperty* qid_p = args.GetProperty(L"qid");
            
            if (filename_p == NULL || regexps_p == NULL || qid_p == NULL)
            {
                throw SCXInternalErrorException(L"missing argument(s) to GetMatchedRows method", SCXSRCLOCATION);
            }
            
            if (filename_p->GetType() != SCXProperty::SCXStringType || 
                regexps_p->GetType() != SCXProperty::SCXArrayType || 
                qid_p->GetType() != SCXProperty::SCXStringType)
            {
                throw SCXInternalErrorException(L"Wrong type of arguments to GetMatchedRows method", SCXSRCLOCATION);
            }
            
            std::wstring filename = filename_p->GetStrValue();
            const std::vector<SCXProperty>& regexps_pv = regexps_p->GetVectorValue();
            std::wstring qid = qid_p->GetStrValue();
            
            SCX_LOGTRACE(m_log, StrAppend(L"SCXLogFileProvider DoInvokeMethod - filename = ", filename));
            SCX_LOGTRACE(m_log, StrAppend(L"SCXLogFileProvider DoInvokeMethod - qid = ", qid));
            SCX_LOGTRACE(m_log, StrAppend(L"SCXLogFileProvider DoInvokeMethod - regexp count = ", regexps_pv.size()));
            
            // Extract and parse the regular expressions
            std::vector<SCXRegexWithIndex> regexps;
            std::wstring invalid_regex(L"");
            
            for (size_t i=0; i<regexps_pv.size(); i++)
            {
                if (regexps_pv[i].GetType() != SCXProperty::SCXStringType)
                {
                    throw SCXInternalErrorException(L"Wrong type of members in regsxps array argument to GetMatchedRows method", SCXSRCLOCATION);
                }
                
                std::wstring regexp = regexps_pv[i].GetStrValue();
                
                SCX_LOGTRACE(m_log, StrAppend(L"SCXLogFileProvider DoInvokeMethod - regexp = ", regexp));
                
                try
                {
                    SCXRegexWithIndex regind;
                    regind.regex = new SCXRegex(regexp);
                    regind.index = i;
                    regexps.push_back(regind);
                }
                catch (SCXInvalidRegexException& e)
                {
                    SCX_LOGWARNING(m_log, StrAppend(L"SCXLogFileProvider DoInvokeMethod - invalid regexp : ", regexp));
                    invalid_regex = StrAppend(StrAppend(invalid_regex, invalid_regex.length()>0?L" ":L""), i);
                }
            }
            
            std::vector<SCXProperty> ans;
            // If any regular expressions with invalid syntax, add special row to result
            if (invalid_regex.length() > 0)
            {
                SCXProperty rowprop(L"row", StrAppend(L"InvalidRegexp;", invalid_regex));
                ans.push_back(rowprop);
            }
            try
            {
                LogFileStreamPositioner positioner(filename, qid, m_persistMedia);
                SCXHandle<std::wfstream> logfile = positioner.GetStream();
                          
                unsigned int rows = 0;
                unsigned int matched_rows = 0;
                
                // Read rows from log file
                while (matched_rows < cMaxMatchedRows && SCXStream::IsGood(*logfile))
                {
                    wstring line;
                    SCXStream::NLF nlf;
                    
                    rows++;
                    
                    SCX_LOGHYSTERICAL(m_log, StrAppend(L"LogFileProvider DoInvokeMethod - Reading row: ", rows));
                    
                    SCXStream::ReadLine(*logfile, line, nlf);
                    
                    // Check line against regular expressions and add to result if any matches
                    std::wstring res(L"");
                    int matches = 0;
                    
                    for (size_t j=0; j<regexps.size(); j++)
                    {
                        if (regexps[j].regex->IsMatch(line))
                        {
                            SCX_LOGHYSTERICAL(m_log, StrAppend(StrAppend(StrAppend(L"LogFileProvider DoInvokeMethod - row: ", rows), 
                                                                         L" Matched regexp: "), regexps[j].index));
                            matches++;
                            res = StrAppend(StrAppend(res, res.length()>0?L" ":L""), regexps[j].index);
                        }
                    }
                    
                    if (matches > 0)
                    {
                        SCXProperty rowprop(L"row", StrAppend(StrAppend(res, L";"), line));
                        ans.push_back(rowprop);
                        matched_rows++;
                    }
                }
                
                // Check if we read all rows, if not add special row to beginning of result
                if (matched_rows >= cMaxMatchedRows && SCXStream::IsGood(*logfile))
                {
                    SCX_LOGINFO(m_log, StrAppend(L"LogFileProvider DoInvokeMethod - Breaking after matching max number of rows : ", cMaxMatchedRows));
                    
                    SCXProperty rowprop(L"row", L"MoreRowsAvailable;true");
                    ans.insert(ans.begin(), rowprop);
                }
                
                positioner.PersistState();
                
                SCXProperty rowsprop(L"rows", ans);
                outargs.AddProperty(rowsprop);

            }
            catch (SCXFilePathNotFoundException& e)
            {
                SCX_LOGWARNING(m_log, StrAppend(L"LogFileProvider DoInvokeMethod - File not found: ", filename).append(e.What()));
            }
            result.SetValue((unsigned int)ans.size());
        }
        else
        {
            throw SCXNotSupportedException(StrAppend(L"Unhandled method name: ", methodname), SCXSRCLOCATION);
        }
    }
    
    /*----------------------------------------------------------------------------*/
    /**
        Set persist media to use.
        Only used for tests

        \param[in]     persistMedia  persistance media to use
    */
    void LogFileProvider::SetPersistMedia(SCXCoreLib::SCXHandle<SCXCoreLib::SCXPersistMedia> persistMedia) 
    { 
        m_persistMedia = persistMedia; 
    }

    /**
        Dump object as string (for logging).
    
        \returns       The object represented as a string suitable for logging.
    
    */
    const std::wstring LogFileProvider::DumpString() const
    {
        return L"LogFileProvider";
    }
}

/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
