/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file

    \brief     RunAs Provider implementation

    \date      07-11-29 12:00:00
*/
/*----------------------------------------------------------------------------*/

#include <scxcorelib/scxcmn.h>
#include <scxcorelib/scxexception.h>

#include <scxproviderlib/scxprovidercapabilities.h>

#include "runasprovider.h"

#include <scxcorelib/scxlog.h>
#include <scxcorelib/stringaid.h>
#include <scxcorelib/scxprocess.h>
#include <scxcorelib/scxfile.h>
#include <scxcorelib/scxfilepath.h>
#include <scxcorelib/scxuser.h>

#include <sstream>
#include <algorithm>

using namespace SCXProviderLib;
using namespace SCXCoreLib;

namespace SCXCore {

   /*----------------------------------------------------------------------------*/
    /**
       Provide CMPI interface for this class

       The class implementation (concrete class) is RunAsProvider and the name of the
       provider in CIM registration terms is SCX_RunAsProvider.
    */
    SCXProviderDef(RunAsProvider, SCX_RunAsProvider)

    /*----------------------------------------------------------------------------*/
    /**
       Default constructor

       \param[in] configurator RunAs provider configurator 

       The Singleton thread lock will be held during this call.
    */
    RunAsProvider::RunAsProvider(SCXCoreLib::SCXHandle<RunAsConfigurator> configurator
                                 /* = new RunAsConfigurator() */) : 
        BaseProvider(L"scx.core.providers.runasprovider"),
        m_Configurator(configurator)
    {
        SCX_LOGTRACE(m_log, L"RunAsProvider constructor");

        EnableProviderUnloading();
    }

    /*----------------------------------------------------------------------------*/
    /**
       Destructor
    */
    RunAsProvider::~RunAsProvider()
    {
        // Do not log here since when this destructor is called the objects neccesary for logging might no longer be alive
    }

    /*----------------------------------------------------------------------------*/
    /**
       Registration of supported capabilities

       Callback from the BaseProvider in which the provider registers the supported
       classes and methods. The registrations should match the contents of the
       MOF files exactly.
    */
    void RunAsProvider::DoInit()
    {
        SCX_LOGTRACE(m_log, L"RunAsProvider::DoInit");

        m_ProviderCapabilities.RegisterCimClass(eSCX_OperatingSystem,
                                                L"SCX_OperatingSystem");
        m_ProviderCapabilities.RegisterCimMethod(eSCX_OperatingSystem, eExecuteCommandMethod,
                                                 L"ExecuteCommand");
        m_ProviderCapabilities.RegisterCimMethod(eSCX_OperatingSystem, eExecuteShellCommandMethod,
                                                 L"ExecuteShellCommand");
        m_ProviderCapabilities.RegisterCimMethod(eSCX_OperatingSystem, eExecuteScriptMethod,
                                                 L"ExecuteScript");
        ParseConfiguration();
    }

    /*----------------------------------------------------------------------------*/
    /**
        Provide a way for pal layer to do cleanup. Stop all threads etc.
    */
    void RunAsProvider::DoCleanup()
    {
        SCX_LOGTRACE(m_log, L"RunAsProvider::DoCleanup");

        m_ProviderCapabilities.Clear();
    }

    /*----------------------------------------------------------------------------*/
    /**
        Execute a command

        \param[in]     command     Command to execute
        \param[out]    resultOut   Result string from stdout
        \param[out]    resultErr   Result string from stderr
        \param[out]    returncode  Return code from command
        \param[in]     timeout     Accepted number of seconds to wait
        \returns       true if command succeeded, else false
        \throws SCXAccessViolationException If execution is prohibited by configuration
    */
    bool RunAsProvider::ExecuteCommand(const std::wstring &command, std::wstring &resultOut, std::wstring &resultErr, int& returncode,
                                       unsigned timeout)
    {
        SCX_LOGTRACE(m_log, L"SCXRunAsProvider ExecuteCommand");

        if ( ! m_Configurator->GetAllowRoot())
        {
            SCXUser currentUser;
            if (currentUser.IsRoot())
            {
                throw SCXAccessViolationException(L"Configuration prohibits execution with user: root", SCXSRCLOCATION);
            }
        }

        std::istringstream processInput;
        std::ostringstream processOutput;
        std::ostringstream processError;

        try
        {
            returncode = SCXCoreLib::SCXProcess::Run(command, processInput, processOutput, processError, timeout * 1000, m_Configurator->GetCWD(), m_Configurator->GetChRootPath());
            SCX_LOGHYSTERICAL(m_log, L"\"" + command + L"\" returned " + StrFrom(returncode));
            resultOut = StrFromMultibyte(processOutput.str());
            SCX_LOGHYSTERICAL(m_log, L"stdout: " + resultOut);
            resultErr = StrFromMultibyte(processError.str());
            SCX_LOGHYSTERICAL(m_log, L"stderr: " + resultErr);
        }
        catch (SCXCoreLib::SCXException& e)
        {
            resultOut = L"";
            resultErr = e.What();
            returncode = -1;
        }

        return (returncode == 0);
    }

    /*----------------------------------------------------------------------------*/
    /**
        Execute a command in the default shell.

        \param[in]     command     Command to execute
        \param[out]    resultOut   Result string from stdout
        \param[out]    resultErr   Result string from stderr
        \param[out]    returncode  Return code from command
        \param[in]     timeout     Accepted number of seconds to wait
        \returns       true if command succeeded, else false
        \throws SCXAccessViolationException If execution is prohibited by configuration
    */
    bool RunAsProvider::ExecuteShellCommand(const std::wstring &command, std::wstring &resultOut, std::wstring &resultErr, int& returncode,
                                            unsigned timeout)
    {
        SCX_LOGTRACE(m_log, L"SCXRunAsProvider ExecuteShellCommand");

        if ( ! m_Configurator->GetAllowRoot())
        {
            SCXUser currentUser;
            if (currentUser.IsRoot())
            {
                throw SCXAccessViolationException(L"Configuration prohibits execution with user: root", SCXSRCLOCATION);
            }
        }

        std::istringstream processInput;
        std::ostringstream processOutput;
        std::ostringstream processError;

        const char* shell = getenv("SHELL");
        std::wstring shellcommand;
        if (0 == shell)
        {
            shellcommand = L"/bin/sh -c \"";
        }
        else
        {
            shellcommand = StrFromMultibyte(shell);
            shellcommand.append(L" -c \"");
        }
        shellcommand.append(command);
        shellcommand.append(L"\"");

        try
        {
            returncode = SCXCoreLib::SCXProcess::Run(shellcommand, processInput, processOutput, processError, timeout * 1000, m_Configurator->GetCWD(), m_Configurator->GetChRootPath());
            SCX_LOGHYSTERICAL(m_log, L"\"" + shellcommand + L"\" returned " + StrFrom(returncode));
            resultOut = StrFromMultibyte(processOutput.str());
            SCX_LOGHYSTERICAL(m_log, L"stdout: " + resultOut);
            resultErr = StrFromMultibyte(processError.str());
            SCX_LOGHYSTERICAL(m_log, L"stderr: " + resultErr);
        }
        catch (SCXCoreLib::SCXException& e)
        {
            resultOut = L"";
            resultErr = e.What();
            returncode = -1;
        }

        return (returncode == 0);
    }

    /*----------------------------------------------------------------------------*/
    /**
        Execute a script

        \param[in]     script      Script to execute
        \param[in]     arguments   Command line arguments to script
        \param[out]    resultOut   Result string from stdout
        \param[out]    resultErr   Result string from stderr
        \param[out]    returncode  Return code from command
        \param[in]     timeout     Accepted number of seconds to wait

        \returns       true if script succeeded, else false
        \throws SCXAccessViolationException If execution is prohibited by configuration    */
    bool RunAsProvider::ExecuteScript(const std::wstring &script, const std::wstring &arguments, std::wstring &resultOut, std::wstring &resultErr, int& returncode,
                                      unsigned timeout)
    {
        SCX_LOGTRACE(m_log, L"SCXRunAsProvider ExecuteScript");

        if ( ! m_Configurator->GetAllowRoot())
        {
            SCXUser currentUser;
            if (currentUser.IsRoot())
            {
                throw SCXAccessViolationException(L"Configuration prohibits execution with user: root", SCXSRCLOCATION);
            }
        }

        std::istringstream processInput;
        std::ostringstream processOutput;
        std::ostringstream processError;

        try
        {
            SCXFilePath scriptfile = SCXFile::CreateTempFile(script);
            SCXFileSystem::Attributes attribs = SCXFileSystem::GetAttributes(scriptfile);
            attribs.insert(SCXFileSystem::eUserExecute);
            SCXFile::SetAttributes(scriptfile, attribs);

            std::wstring command(scriptfile.Get());
            command.append(L" ").append(arguments);
            returncode = SCXCoreLib::SCXProcess::Run(command, processInput, processOutput, processError, timeout * 1000, m_Configurator->GetCWD(), m_Configurator->GetChRootPath());
            SCXFile::Delete(scriptfile);

            SCX_LOGHYSTERICAL(m_log, L"\"" + command + L"\" returned " + StrFrom(returncode));
            resultOut = StrFromMultibyte(processOutput.str());
            SCX_LOGHYSTERICAL(m_log, L"stdout: " + resultOut);
            resultErr = StrFromMultibyte(processError.str());
            SCX_LOGHYSTERICAL(m_log, L"stderr: " + resultErr);
        }
        catch (SCXCoreLib::SCXException& e)
        {
            resultOut = L"";
            resultErr = e.What();
            returncode = -1;
        }

        return (returncode == 0);
    }

    /*----------------------------------------------------------------------------*/
    /**
        Invoke a method on an instance

        \param[in]     callContext Keys indicating instance to execute method on
        \param[in]     methodname  Name of method called
        \param[in]     args        Arguments provided for method call
        \param[out]    outargs     Output arguments
        \param[out]    result      Result value
    */
    void RunAsProvider::DoInvokeMethod(
        const SCXCallContext& callContext,
        const std::wstring& methodname,
        const SCXArgs& args,
        SCXArgs& outargs,
        SCXProperty& result)
    {
        SCX_LOGTRACE(m_log, L"SCXRunAsProvider DoInvokeMethod");

        SupportedCimMethods cimmethod = static_cast<SupportedCimMethods>(m_ProviderCapabilities.GetCimMethodId(callContext.GetObjectPath(), methodname));

        if (cimmethod == eExecuteCommandMethod || cimmethod == eExecuteShellCommandMethod)
        {
            const SCXProperty* command = args.GetProperty(L"Command");
            const SCXProperty* timeout = args.GetProperty(L"timeout");
            int returncode = 0;

            if (command == NULL || timeout == NULL)
            {
                throw SCXInternalErrorException(L"missing arguments to ExecuteCommand method", SCXSRCLOCATION);
            }

            if (command->GetType() != SCXProperty::SCXStringType ||
                timeout->GetType() != SCXProperty::SCXUIntType)
            {
                throw SCXInternalErrorException(L"Wrong type of arguments to ExecuteCommand method", SCXSRCLOCATION);
            }

            std::wstring return_out, return_err;
            bool cmdok;
            if (cimmethod == eExecuteCommandMethod)
            {
                cmdok = ExecuteCommand(command->GetStrValue(), return_out, return_err, returncode, timeout->GetUIntValue());
            }
            else
            {
                cmdok = ExecuteShellCommand(command->GetStrValue(), return_out, return_err, returncode, timeout->GetUIntValue());
            }

            result.SetValue(cmdok);

            SCXProperty ret_prop(L"ReturnCode", returncode);
            outargs.AddProperty(ret_prop);
            SCXProperty stdout_prop(L"StdOut", return_out);
            outargs.AddProperty(stdout_prop);
            SCXProperty stderr_prop(L"StdErr", return_err);
            outargs.AddProperty(stderr_prop);
        }
        else if (cimmethod == eExecuteScriptMethod)
        {
            const SCXProperty* script = args.GetProperty(L"Script");
            const SCXProperty* arguments = args.GetProperty(L"Arguments");
            const SCXProperty* timeout = args.GetProperty(L"timeout");
            int returncode = 0;

            if (script == NULL)
            {
                throw SCXInternalErrorException(L"missing arguments to ExecuteScript method", SCXSRCLOCATION);
            }

            if (script->GetType() != SCXProperty::SCXStringType)
            {
                throw SCXInternalErrorException(L"Wrong type of arguments to ExecuteScript method", SCXSRCLOCATION);
            }

            if (arguments == NULL || timeout == NULL)
            {
                throw SCXInternalErrorException(L"missing arguments to ExecuteScript method", SCXSRCLOCATION);
            }

            if (arguments->GetType() != SCXProperty::SCXStringType ||
                timeout->GetType() != SCXProperty::SCXUIntType)
            {
                throw SCXInternalErrorException(L"Wrong type of arguments to ExecuteScript method", SCXSRCLOCATION);
            }

            std::wstring return_out, return_err;
            std::wstring strScript = script->GetStrValue();
            // before Peg 2.9, open wsman removed '\r' characters
            std::wstring::size_type pos_slash_r = strScript.find( '\r' );

            while ( std::wstring::npos != pos_slash_r ){
                strScript.erase( pos_slash_r, 1 );
                pos_slash_r = strScript.find( '\r' );
            }
            
            bool cmdok = ExecuteScript(strScript, arguments->GetStrValue(), return_out, return_err, returncode, timeout->GetUIntValue());

            result.SetValue(cmdok);

            SCXProperty ret_prop(L"ReturnCode", returncode);
            outargs.AddProperty(ret_prop);
            SCXProperty stdout_prop(L"StdOut", return_out);
            outargs.AddProperty(stdout_prop);
            SCXProperty stderr_prop(L"StdErr", return_err);
            outargs.AddProperty(stderr_prop);
        }
        else
        {
            throw SCXInternalErrorException(StrAppend(L"Unhandled method name: ", methodname), SCXSRCLOCATION);
        }
    }

    /*----------------------------------------------------------------------------*/
    /**
       Parses the configuratin file.

       All parse errors are logged. If configuraion cannot be parsed, default
       configuration is used.

    */
    void RunAsProvider::ParseConfiguration()
    {
        m_Configurator->Parse();
    }
}

/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
