/*
**==============================================================================
**
** Microsoft Public License (Ms-PL)
** 
** This license governs use of the accompanying software. If you use the 
** software, you accept this license. If you do not accept the license, do not 
** use the software.
**
** 1. Definitions
** The terms "reproduce," "reproduction," "derivative works," and 
** "distribution" have the same meaning here as under U.S. copyright law.
** A "contribution" is the original software, or any additions or changes to 
** the software.
** A "contributor" is any person that distributes its contribution under this 
** license.
** "Licensed patents" are a contributor's patent claims that read directly on 
** its contribution.
**
** 2. Grant of Rights
** (A) Copyright Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free copyright license to reproduce its
** contribution, prepare derivative works of its contribution, and distribute 
** its contribution or any derivative works that you create.
** (B) Patent Grant- Subject to the terms of this license, including the 
** license conditions and limitations in section 3, each contributor grants you
** a non-exclusive, worldwide, royalty-free license under its licensed patents
** to make, have made, use, sell, offer for sale, import, and/or otherwise 
** dispose of its contribution in the software or derivative works of the 
** contribution in the software.
**
** 3. Conditions and Limitations
** (A) No Trademark License- This license does not grant you rights to use any 
** contributors' name, logo, or trademarks.
** (B) If you bring a patent claim against any contributor over patents that 
** you claim are infringed by the software, your patent license from such 
** contributor to the software ends automatically.
** (C) If you distribute any portion of the software, you must retain all 
** copyright, patent, trademark, and attribution notices that are present in 
** the software.
** (D) If you distribute any portion of the software in source code form, you 
** may do so only under this license by including a complete copy of this 
** license with your distribution. If you distribute any portion of the 
** software in compiled or object code form, you may only do so under a license
** that complies with this license.
** (E) The software is licensed "as-is." You bear the risk of using it. The 
** contributors give no express warranties, guarantees or conditions. You may 
** have additional consumer rights under your local laws which this license 
** cannot change. To the extent permitted under your local laws, the 
** contributors exclude the implied warranties of merchantability, fitness for 
** a particular purpose and non-infringement.
**
**==============================================================================
*/
/**
    \file

    \brief     RunAs Provider

    \date      07-11-29 12:00:00
*/
/*----------------------------------------------------------------------------*/
#ifndef RUNASPROVIDER_H
#define RUNASPROVIDER_H

#include <string>

#include <scxcorelib/scxlog.h>
#include <scxcorelib/scxhandle.h>
#include <scxcorelib/scxfilepath.h>
#include <scxproviderlib/cmpibase.h>

#include "scxrunasconfigurator.h"

namespace SCXCore
{
    /*----------------------------------------------------------------------------*/
    /**
       RunAs provider

       Concrete instance of the CMPI BaseProvider implementing a method
       for executing commands

       A provider-specific thread lock will be held at each call to
       the Do* methods, so this implementation class does not need
       to worry about that.
    */
    class RunAsProvider : public SCXProviderLib::BaseProvider
    {
    public:
        /**
           Default constructor
           \param[in] configurator RunAs provider configurator 
           The Singleton thread lock will be held during this call.
        */
        RunAsProvider(SCXCoreLib::SCXHandle<RunAsConfigurator> configurator
                      = SCXCoreLib::SCXHandle<RunAsConfigurator>(new RunAsConfigurator()));
        ~RunAsProvider();

    protected:
        //! The set of CIM classes this provider supports
        enum SupportedCimClasses {
            eSCX_OperatingSystem
        };

        //! The CIM methods this provider supports
        enum SupportedCimMethods {
            eExecuteCommandMethod,
            eExecuteShellCommandMethod,
            eExecuteScriptMethod
        };

        // Overrides from the base class with relevant implementations
        virtual void DoInit();
        virtual void DoCleanup();

        virtual void DoInvokeMethod(const SCXProviderLib::SCXCallContext& callContext,
                                    const std::wstring& methodname, const SCXProviderLib::SCXArgs& args,
                                    SCXProviderLib::SCXArgs& outargs, SCXProviderLib::SCXProperty& result);

    private:
        bool ExecuteCommand(const std::wstring &command, std::wstring &resultOut,
                            std::wstring &resultErr, int& returncode, unsigned timeout = 0);

        bool ExecuteShellCommand(const std::wstring &command, std::wstring &resultOut,
                                 std::wstring &resultErr, int& returncode, unsigned timeout = 0);

        bool ExecuteScript(const std::wstring &script, const std::wstring &arguments,
                           std::wstring &resultOut, std::wstring &resultErr,
                           int& returncode, unsigned timeout = 0);

        void ParseConfiguration();
        
        //! Configurator.
        SCXCoreLib::SCXHandle<RunAsConfigurator> m_Configurator;
    };
}

#endif /* RUNASPROVIDER_H */
/*----------------------------E-N-D---O-F---F-I-L-E---------------------------*/
